from __future__ import annotations

import os
import uuid
from os import path as osp
from pathlib import Path
from typing import Dict, List, Optional

import supervisely.convert.pointcloud_episodes.nuscenes_conv.nuscenes_helper as helpers
import supervisely.io.fs as fs
from supervisely._utils import is_development
from supervisely.annotation.obj_class import ObjClass
from supervisely.annotation.tag_meta import TagMeta, TagValueType
from supervisely.api.api import Api, ApiField
from supervisely.api.dataset_api import DatasetInfo
from supervisely.convert.base_converter import AvailablePointcloudConverters
from supervisely.convert.pointcloud_episodes.pointcloud_episodes_converter import (
    PointcloudEpisodeConverter,
)
from supervisely.geometry.cuboid_3d import Cuboid3d
from supervisely.pointcloud_annotation.pointcloud_episode_annotation import (
    PointcloudEpisodeAnnotation,
)
from supervisely.pointcloud_annotation.pointcloud_episode_frame import (
    PointcloudEpisodeFrame,
)
from supervisely.pointcloud_annotation.pointcloud_episode_frame_collection import (
    PointcloudEpisodeFrameCollection,
)
from supervisely.pointcloud_annotation.pointcloud_episode_object import (
    PointcloudEpisodeObject,
)
from supervisely.pointcloud_annotation.pointcloud_episode_object_collection import (
    PointcloudEpisodeObjectCollection,
)
from supervisely.pointcloud_annotation.pointcloud_episode_tag_collection import (
    PointcloudEpisodeTagCollection,
)
from supervisely.pointcloud_annotation.pointcloud_figure import PointcloudFigure
from supervisely.project.project_meta import ProjectMeta
from supervisely.sly_logger import logger
from supervisely.tiny_timer import TinyTimer
from supervisely.video_annotation.key_id_map import KeyIdMap


class NuscenesEpisodesConverter(PointcloudEpisodeConverter):
    """Converter for NuScenes pointcloud episodes format."""

    _nuscenes: "NuScenes" = None  # type: ignore
    _custom_data: Dict = {}

    def __str__(self) -> str:
        return AvailablePointcloudConverters.NUSCENES

    def validate_format(self) -> bool:
        try:
            from nuscenes import NuScenes
        except ImportError:
            logger.warning("Please, run 'pip install nuscenes-devkit' to import NuScenes data.")
            return False

        table_json_filenames = [f"{name}.json" for name in helpers.TABLE_NAMES]

        def _contains_tables(dir_path: str) -> bool:
            return all(fs.file_exists(osp.join(dir_path, table)) for table in table_json_filenames)

        def _filter_fn(path):
            has_tables = False
            for p in os.scandir(path):
                if p.is_dir() and _contains_tables(p.path):
                    has_tables = True
                    break
            return has_tables and (Path(path) / "samples").exists()

        input_path = next((d for d in fs.dirs_filter(self._input_data, _filter_fn)), None)
        if input_path is None:
            return False

        ann_dir = next((d for d in fs.dirs_filter(input_path, _contains_tables)), None)
        if ann_dir is None:
            return False

        version = osp.basename(ann_dir)
        try:
            t = TinyTimer()
            self._nuscenes: NuScenes = NuScenes(version=version, dataroot=input_path, verbose=False)
            logger.debug(f"NuScenes initialization took {t.get_sec():.3f} sec")
        except Exception as e:
            logger.debug(f"Failed to initialize NuScenes: {e}")
            return False

        self._custom_data["nuscenes_version"] = version
        self._custom_data["dataroot"] = input_path
        return True

    def to_supervisely(
        self,
        scene_samples: Dict[str, helpers.Sample],
        meta: ProjectMeta,
        renamed_classes: dict = {},
        renamed_tags: dict = {},
    ) -> PointcloudEpisodeAnnotation:
        token_to_obj = {}
        frames = []
        tags = []
        frame_idx_to_scene_sample_token = {}
        if "frame_token_map" not in self._custom_data:
            self._custom_data["frame_token_map"] = {}
        for sample_i, (token, sample) in enumerate(scene_samples.items()):
            figures = []
            for obj in sample.anns:
                ann_token = uuid.UUID(obj.token)
                instance_token = obj.instance_token
                class_name = obj.category
                parent_obj_token = obj.parent_token
                parent_object = None
                if parent_obj_token == "":
                    # * Create a new object
                    obj_class_name = renamed_classes.get(class_name, class_name)
                    obj_class = meta.get_obj_class(obj_class_name)
                    obj_tags = None  # ! TODO: fix tags
                    pcd_ep_obj = PointcloudEpisodeObject(
                        obj_class, obj_tags, uuid.UUID(instance_token)
                    )
                    # * Assign the object to the starting token
                    token_to_obj[instance_token] = pcd_ep_obj
                    parent_object = pcd_ep_obj
                else:
                    # * -> Figure has a parent object, get it
                    token_to_obj[instance_token] = token_to_obj[parent_obj_token]
                    parent_object = token_to_obj[parent_obj_token]
                geom = obj.to_supervisely()
                pcd_figure = PointcloudFigure(parent_object, geom, sample_i, ann_token)
                figures.append(pcd_figure)
                frame_idx_to_scene_sample_token[sample_i] = token
            frame = PointcloudEpisodeFrame(sample_i, figures)
            frames.append(frame)
        tag_collection = PointcloudEpisodeTagCollection(tags) if len(tags) > 0 else None
        self._custom_data["frame_token_map"][self._current_ds_id] = frame_idx_to_scene_sample_token
        key_uuid = uuid.UUID(token)
        return PointcloudEpisodeAnnotation(
            len(frames),
            PointcloudEpisodeObjectCollection(list(set(token_to_obj.values()))),
            PointcloudEpisodeFrameCollection(frames),
            tag_collection,
            key=key_uuid,
        )

    def upload_dataset(self, api: Api, dataset_id: int, batch_size: int = 1, log_progress=True):
        from nuscenes import NuScenes  # pylint: disable=import-error

        nuscenes: NuScenes = self._nuscenes
        key_id_map = KeyIdMap()

        tag_metas = [TagMeta(attr["name"], TagValueType.NONE) for attr in nuscenes.attribute]
        obj_classes = {}
        for category in nuscenes.category:
            color = nuscenes.colormap[category["name"]]
            description = helpers.trim_description(category.get("description", ""))
            token = category["token"]
            obj_classes[token] = ObjClass(
                category["name"], Cuboid3d, color, description=description
            )

        self._custom_data["classes_token_map"] = {k: v.name for k, v in obj_classes.items()}

        self._meta = ProjectMeta(list(obj_classes.values()), tag_metas)
        meta, renamed_classes, renamed_tags = self.merge_metas_with_conflicts(api, dataset_id)

        dataset_info = api.dataset.get_info_by_id(dataset_id)
        scene_name_to_dataset: Dict[str, DatasetInfo] = {}

        scene_names = [scene["name"] for scene in nuscenes.scene]
        scene_cnt = len(scene_names)
        total_sample_cnt = sum([scene["nbr_samples"] for scene in nuscenes.scene])

        multiple_scenes = len(scene_names) > 1
        if multiple_scenes:
            logger.info(f"Found {scene_cnt} scenes ({total_sample_cnt} samples) in the input data.")
            # * Create a nested dataset for each scene
            for name in scene_names:
                ds = api.dataset.create(
                    dataset_info.project_id,
                    name,
                    change_name_if_conflict=True,
                    parent_id=dataset_id,
                )
                scene_name_to_dataset[name] = ds
        else:
            scene_name_to_dataset[scene_names[0]] = dataset_info

        if log_progress:
            progress, progress_cb = self.get_progress(
                total_sample_cnt, "Converting episode scenes..."
            )
        else:
            progress_cb = None

        for scene in nuscenes.scene:
            current_dataset_id = scene_name_to_dataset[scene["name"]].id
            self._current_ds_id = current_dataset_id

            log = nuscenes.get("log", scene["log_token"])
            sample_token = scene["first_sample_token"]

            # * Extract scene's samples
            scene_samples: Dict[str, helpers.Sample] = {}
            for i in range(scene["nbr_samples"]):
                sample = nuscenes.get("sample", sample_token)
                lidar_path, boxes, _ = nuscenes.get_sample_data(sample["data"]["LIDAR_TOP"])
                if not osp.exists(lidar_path):
                    logger.warning(f'Scene "{scene["name"]}" has no LIDAR data.')
                    continue

                timestamp = sample["timestamp"]
                anns = []
                for box, name, inst_token in helpers.Sample.generate_boxes(nuscenes, boxes):
                    current_instance_token = inst_token["token"]
                    parent_token = inst_token["prev"]

                    ann = nuscenes.get("sample_annotation", current_instance_token)
                    category = ann["category_name"]
                    attributes = [
                        nuscenes.get("attribute", attr)["name"] for attr in ann["attribute_tokens"]
                    ]
                    visibility = nuscenes.get("visibility", ann["visibility_token"])["level"]
                    ann_token = ann["token"]

                    ann = helpers.AnnotationObject(
                        name=name,
                        bbox=box,
                        token=ann_token,
                        instance_token=current_instance_token,
                        parent_token=parent_token,
                        category=category,
                        attributes=attributes,
                        visibility=visibility,
                    )
                    anns.append(ann)

                # get camera data
                sample_data = nuscenes.get("sample_data", sample["data"]["LIDAR_TOP"])
                cal_sensor = nuscenes.get(
                    "calibrated_sensor", sample_data["calibrated_sensor_token"]
                )
                ego_pose = nuscenes.get("ego_pose", sample_data["ego_pose_token"])

                camera_data = [
                    helpers.CamData(nuscenes, sensor, token, cal_sensor, ego_pose)
                    for sensor, token in sample["data"].items()
                    if sensor.startswith("CAM")
                ]
                sample_token = sample["token"]
                scene_samples[sample_token] = helpers.Sample(
                    timestamp, lidar_path, anns, camera_data
                )
                sample_token = sample["next"]

            # * Convert and upload pointclouds
            frame_to_pointcloud_ids = {}
            for idx, sample in enumerate(scene_samples.values()):
                pcd_path = sample.convert_lidar_to_supervisely()

                pcd_name = fs.get_file_name_with_ext(pcd_path)
                pcd_meta = {
                    "frame": idx,
                    "vehicle": log["vehicle"],
                    "date": log["date_captured"],
                    "location": log["location"],
                    "description": scene["description"],
                }
                info = api.pointcloud_episode.upload_path(
                    current_dataset_id, pcd_name, pcd_path, pcd_meta
                )
                fs.silent_remove(pcd_path)

                pcd_id = info.id
                frame_to_pointcloud_ids[idx] = pcd_id

                # * Upload related images
                image_jsons = []
                camera_names = []
                for img_path, rimage_info in [
                    data.get_info(sample.timestamp) for data in sample.cam_data
                ]:
                    img = api.pointcloud_episode.upload_related_image(img_path)
                    image_jsons.append(
                        {
                            ApiField.ENTITY_ID: pcd_id,
                            ApiField.NAME: rimage_info[ApiField.NAME],
                            ApiField.HASH: img,
                            ApiField.META: rimage_info[ApiField.META],
                        }
                    )
                    camera_names.append(rimage_info[ApiField.META]["deviceId"])
                if len(image_jsons) > 0:
                    api.pointcloud_episode.add_related_images(image_jsons, camera_names)

                if log_progress:
                    progress_cb(1)

            # * Convert and upload annotations
            pcd_ann = self.to_supervisely(scene_samples, meta, renamed_classes, renamed_tags)

            try:
                api.pointcloud_episode.annotation.append(
                    current_dataset_id, pcd_ann, frame_to_pointcloud_ids, key_id_map=key_id_map
                )
                logger.info(f"Dataset ID:{current_dataset_id} has been successfully uploaded.")
            except Exception as e:
                error_msg = getattr(getattr(e, "response", e), "text", str(e))
                logger.warning(
                    f"Failed to upload annotation for scene: {scene['name']}. Message: {error_msg}"
                )
        key_id_map = key_id_map.to_dict()
        key_id_map.pop("tags")
        key_id_map.pop("videos")
        self._custom_data["key_id_map"] = key_id_map

        project_id = dataset_info.project_id
        current_custom_data = api.project.get_custom_data(project_id)
        current_custom_data.update(self._custom_data)
        api.project.update_custom_data(project_id, current_custom_data)

        if log_progress:
            if is_development():
                progress.close()



from typing import Dict

from supervisely.nn.benchmark.semantic_segmentation.base_vis_metric import (
    SemanticSegmVisMetric,
)
from supervisely.nn.benchmark.visualization.widgets import ChartWidget, MarkdownWidget


class FrequentlyConfused(SemanticSegmVisMetric):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.clickable = True
        self._keypair_sep = "-"

    @property
    def md(self) -> MarkdownWidget:
        if self.is_empty:
            text = self.vis_texts.markdown_frequently_confused_empty
        else:
            text = self.vis_texts.markdown_frequently_confused
        return MarkdownWidget("frequently_confused", "Frequently Confused Classes", text=text)

    @property
    def chart(self) -> ChartWidget:
        chart = ChartWidget("frequently_confused", self.get_figure())
        chart.set_click_data(
            self.explore_modal_table.id,
            self.get_click_data(),
            chart_click_extra="'getKey': (payload) => `${payload.points[0].x}`, 'keySeparator': '-',",
        )
        return chart

    @property
    def is_empty(self) -> bool:
        probs, indexes_2d = self.eval_result.mp.frequently_confused
        return len(probs) == 0

    def get_figure(self):
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()

        # Frequency of Confused Classes figure
        probs, indexes_2d = self.eval_result.mp.frequently_confused
        confused_classes = []
        for idx in indexes_2d:
            gt_idx, pred_idx = idx[0], idx[1]
            gt_class = self.eval_result.mp.eval_data.index[gt_idx]
            pred_class = self.eval_result.mp.eval_data.index[pred_idx]
            confused_classes.append(f"{gt_class}-{pred_class}")

        fig = go.Figure()
        fig.add_trace(
            go.Bar(
                x=confused_classes,
                y=probs,
                orientation="v",
                text=probs,
                marker=dict(color=probs, colorscale="Reds"),
            )
        )
        fig.update_traces(hovertemplate="Class Pair: %{x}<br>Probability: %{y:.2f}<extra></extra>")
        fig.update_layout(
            xaxis_title="Class Pair",
            yaxis_title="Probability",
            yaxis_range=[0, max(probs) + 0.1],
            yaxis=dict(showticklabels=False),
            font=dict(size=24),
            width=1000 if len(confused_classes) > 10 else 600,
        )

        return fig

    def get_click_data(self) -> Dict:
        if self.is_empty:
            return
        res = dict(projectMeta=self.eval_result.pred_project_meta.to_json())

        res["layoutTemplate"] = [None, None, None]
        res["clickData"] = {}

        _, class_names = self.eval_result.mp.confusion_matrix
        _, indexes_2d = self.eval_result.mp.frequently_confused
        for idx in indexes_2d:
            gt_idx, pred_idx = idx[0], idx[1]
            gt_key = class_names[gt_idx]
            pred_key = class_names[pred_idx]
            key = f"{gt_key}{self._keypair_sep}{pred_key}"

            res["clickData"][key] = {}
            res["clickData"][key]["imagesIds"] = []
            idx_key = str(gt_idx) + "_" + str(pred_idx)
            for name in self.eval_result.mp.cmat_cell_img_names.get(idx_key, []):
                gt_img_id = self.eval_result.images_map[name]
                pred_img_id = self.eval_result.matched_pair_data[gt_img_id].pred_image_info.id
                res["clickData"][key]["imagesIds"].append(pred_img_id)

            title = f"Confused classes. GT: '{gt_key}' ― Predicted: '{pred_key}'"
            res["clickData"][key]["title"] = title
        return res



from __future__ import annotations
from typing import List, Dict, Optional, Union, Callable
from supervisely.app.widgets import Widget, Select, Button, Flexbox
from supervisely.sly_logger import logger


class SelectCudaDevice(Widget):
    """
    A widget for selecting a CUDA device.

    This widget allows to select a CUDA device (and optional CPU device) from a list of detected devices on the machine.
    It displays the devices along with their reserved/total RAM values.

    :param get_list_on_init: Whether to retrieve and display the list of CUDA devices upon initialization.
    :type get_list_on_init: bool, optional
    :param sort_by_free_ram: Whether to sort the CUDA devices by their available free RAM.
    :type sort_by_free_ram: bool, optional
    :param include_cpu_option: Whether to include an option to select the CPU in the device list.
    :type include_cpu_option: bool, optional
    """

    def __init__(
        self,
        get_list_on_init: Optional[bool] = True,
        sort_by_free_ram: Optional[bool] = False,
        include_cpu_option: Optional[bool] = False,
        widget_id: str = None,
    ):
        self._select = Select([])
        self._refresh_button = Button(
            text="", button_type="text", button_size="large", icon="zmdi zmdi-refresh"
        )
        self._refresh_button.click(self.refresh)
        self._content = Flexbox([self._select, self._refresh_button])

        self._sort_by_free_ram = sort_by_free_ram
        self._include_cpu_option = include_cpu_option
        if get_list_on_init:
            self.refresh()

        super().__init__(widget_id=widget_id, file_path=__file__)

    def refresh(self) -> None:
        """Refreshes the list of available CUDA devices and updates the selector's items.

        :return: None
        """
        cuda_devices = self._get_gpu_infos(self._sort_by_free_ram)
        if cuda_devices is None:
            if self._include_cpu_option is True:
                self._select.set([Select.Item(value="cpu", label="CPU")])
            else:
                self._select.set([Select.Item(None, "No devices available")])
            self._select.disable()
            return
        items = [
            Select.Item(
                value=info["device_idx"],
                label=device,
                right_text=info["right_text"],
            )
            for device, info in cuda_devices.items()
        ]

        if self._include_cpu_option:
            items.append(Select.Item(value="cpu", label="CPU"))
        self._select.set(items)

    def _get_gpu_infos(
        self, sort_by_free_ram: bool
    ) -> Optional[Dict[str, Dict[str, Union[str, int]]]]:
        """
        Gets list of all available CUDA devices on current machine and returns a dictionary designed to construct selector widget.
        Optionally, could be sorted by free ram amount.

        :param sort_by_free_ram: Whether to sort the CUDA devices dictionary by device's available free RAM.
        :type sort_by_free_ram: bool, optional
        :return: The dictionary with full device name as a key, and a dictionary with device's info as a value.
        :rtype: Optional[Dict[str, Dict[str, Union[str, int]]]]
        """
        try:
            from torch import cuda
        except ImportError as ie:
            logger.warn(
                "Unable to import Torch. Please, run 'pip install torch' to resolve the issue.",
                extra={"error message": str(ie)},
            )
            return

        devices = {}
        try:
            cuda.init()
            if not cuda.is_available():
                raise RuntimeError("CUDA is not available")
        except Exception as e:
            logger.warn(f"Failed to initialize CUDA: {e}")
            return
        try:
            for idx in range(cuda.device_count()):
                current_device = f"cuda:{idx}"
                full_device_name = f"{cuda.get_device_name(idx)} ({current_device})"
                free_mem, total_mem = cuda.mem_get_info(current_device)

                convert_to_gb = lambda number: round(number / 1024**3, 1)
                right_text = (
                    f"{convert_to_gb(total_mem - free_mem)} GB / {convert_to_gb(total_mem)} GB"
                )

                devices[full_device_name] = {
                    "device_idx": current_device,
                    "right_text": right_text,
                    "free": free_mem,
                }
            if sort_by_free_ram:
                devices = dict(
                    sorted(devices.items(), key=lambda item: item[1]["free"], reverse=True)
                )
        except Exception as e:
            logger.warning(repr(e))
        finally:
            return devices

    def get_json_data(self) -> Dict:
        """Get the JSON data of the widget.

        :return: The JSON data.
        :rtype: Dict
        """
        return {}

    def get_json_state(self) -> Dict:
        """Get the JSON state of the widget.

        :return: The JSON data.
        :rtype: Dict
        """
        return {}

    def value_changed(self, func: Callable) -> Callable:
        """
        This decorator function allows to handle the value change event of the widget.

        :return: Callable
        """

        @self._select.value_changed
        def value_cb(value):
            func(value)

        return value_cb

    def get_device(self) -> Optional[str]:
        """Gets the currently selected device.
        This method returns the value of the currently selected device.

        :return: The value of the selected device (e.g. 'cuda:0', 'cpu', etc.), or None if no device is selected.
        :rtype: Optional[str]
        """
        return self._select.get_value()

    def set_device(self, value: str) -> None:
        """Sets the currently selected device.

        This method updates the selector with the provided device value.

        :param value: The value of the device to be selected (e.g. 'cuda:0', 'cpu', etc.).
        :type value: str
        :return: None
        """
        return self._select.set_value(value)

    def disable(self) -> None:
        """Disables the widget.

        This method disables the widget and grays out the selector.

        :return: None
        """
        self._disabled = True
        self._select.disable()
        self._refresh_button.disable()

    def enable(self) -> None:
        """Enables the widget.

        This method enables the widget and makes the selector clickable.

        :return: None
        """
        self._disabled = False
        self._select.enable()
        self._refresh_button.enable()



# coding: utf-8
# isort: skip_file
import pkg_resources  # isort: skip
import os

try:
    __version__ = pkg_resources.require("supervisely")[0].version
except TypeError as e:
    __version__ = "development"


class _ApiProtoNotAvailable:
    """Placeholder class that raises an error when accessing any attribute"""

    def __getattr__(self, name):
        from supervisely.app.v1.constants import PROTOBUF_REQUIRED_ERROR

        raise ImportError(f"Cannot access `api_proto.{name}` : " + PROTOBUF_REQUIRED_ERROR)

    def __bool__(self):
        return False

    def __repr__(self):
        return "<api_proto: not available - install supervisely[agent] to enable>"


from supervisely.sly_logger import (
    logger,
    ServiceType,
    EventType,
    add_logger_handler,
    add_default_logging_into_file,
    get_task_logger,
    change_formatters_default_values,
    LOGGING_LEVELS,
)

from supervisely.function_wrapper import (
    main_wrapper,
    function_wrapper,
    catch_silently,
    function_wrapper_nofail,
    function_wrapper_external_logger,
)

from supervisely.io import fs
from supervisely.io import env
from supervisely.io import json
from supervisely.io import network_exceptions
from supervisely.io.fs_cache import FileCache

from supervisely.imaging import image

# legacy
# import supervisely.imaging.video as imagevideo
from supervisely.imaging import color

from supervisely.task.paths import TaskPaths

from supervisely.task.progress import (
    epoch_float,
    Progress,
    report_import_finished,
    report_dtl_finished,
    report_dtl_verification_finished,
    report_metrics_training,
    report_metrics_validation,
    report_inference_finished,
)


import supervisely.project as project
import supervisely.api.constants as api_constants
from supervisely.project import read_project, get_project_class
from supervisely.project.download import download, download_async, download_fast
from supervisely.project.upload import upload
from supervisely.project.project import (
    Project,
    OpenMode,
    download_project,
    read_single_project,
    upload_project,
    Dataset,
)
from supervisely.project.project_meta import ProjectMeta

from supervisely.annotation.annotation import ANN_EXT, Annotation
from supervisely.annotation.label import Label
from supervisely.annotation.obj_class import ObjClass, ObjClassJsonFields
from supervisely.annotation.obj_class_collection import ObjClassCollection
from supervisely.annotation.tag_meta import TagMeta, TagValueType, TagApplicableTo
from supervisely.annotation.tag import Tag
from supervisely.annotation.tag_collection import TagCollection
from supervisely.annotation.tag_meta_collection import TagMetaCollection

from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.cuboid import Cuboid
from supervisely.geometry.point import Point
from supervisely.geometry.point_location import PointLocation
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.polyline import Polyline
from supervisely.geometry.rectangle import Rectangle
from supervisely.geometry.mask_3d import Mask3D
from supervisely.geometry.any_geometry import AnyGeometry
from supervisely.geometry.graph import GraphNodes, Node
from supervisely.geometry.multichannel_bitmap import MultichannelBitmap
from supervisely.geometry.alpha_mask import AlphaMask
from supervisely.geometry.cuboid_2d import Cuboid2d
from supervisely.geometry.oriented_bbox import OrientedBBox

from supervisely.geometry.helpers import geometry_to_bitmap
from supervisely.geometry.helpers import deserialize_geometry

from supervisely.export.pascal_voc import save_project_as_pascal_voc_detection

from supervisely.metric.metric_base import MetricsBase
from supervisely.metric.projects_applier import MetricProjectsApplier


from supervisely.metric.iou_metric import IoUMetric
from supervisely.metric.confusion_matrix_metric import ConfusionMatrixMetric
from supervisely.metric.precision_recall_metric import PrecisionRecallMetric
from supervisely.metric.classification_metrics import ClassificationMetrics
from supervisely.metric.map_metric import MAPMetric

from supervisely.worker_api.agent_api import AgentAPI
from supervisely.worker_api.chunking import (
    ChunkSplitter,
    ChunkedFileWriter,
    ChunkedFileReader,
)

# Global import of api_proto works only if protobuf is installed and compatible
# Otherwise, we use a placeholder that raises an error when accessed
try:
    import supervisely.worker_proto.worker_api_pb2 as api_proto
except Exception:
    api_proto = _ApiProtoNotAvailable()


from supervisely.api.api import Api, UserSession, ApiContext
from supervisely.api import api
from supervisely.api.task_api import WaitingTimeExceeded, TaskFinishedWithError
from supervisely.project.project_type import ProjectType
from supervisely.project.project_settings import ProjectSettings
from supervisely.api.report_api import NotificationType
from supervisely.api.image_api import ImageInfo
from supervisely.api.dataset_api import DatasetInfo
from supervisely.api.project_api import ProjectInfo
from supervisely.api.workspace_api import WorkspaceInfo
from supervisely.api.team_api import TeamInfo
from supervisely.api.entity_annotation.figure_api import FigureInfo
from supervisely.api.app_api import WorkflowSettings, WorkflowMeta
from supervisely.api.entities_collection_api import EntitiesCollectionInfo

from supervisely.cli import _handle_creds_error_to_console

from supervisely._utils import (
    rand_str,
    batched,
    batched_iter,
    get_bytes_hash,
    generate_names,
    ENTERPRISE,
    COMMUNITY,
    _dprint,
    take_with_default,
    get_string_hash,
    is_development,
    is_production,
    is_debug_with_sly_net,
    compress_image_url,
    get_datetime,
    get_readable_datetime,
    generate_free_name,
    setup_certificates,
    is_community,
    run_coroutine,
)

import supervisely._utils as utils
from supervisely.tiny_timer import TinyTimer

from supervisely.aug import aug
from supervisely.video_annotation.key_id_map import KeyIdMap

from supervisely.video_annotation.video_annotation import VideoAnnotation
from supervisely.video_annotation.video_object import VideoObject
from supervisely.video_annotation.video_object_collection import VideoObjectCollection
from supervisely.video_annotation.video_figure import VideoFigure
from supervisely.video_annotation.frame import Frame
from supervisely.video_annotation.frame_collection import FrameCollection
from supervisely.video_annotation.video_tag import VideoTag
from supervisely.video_annotation.video_tag_collection import VideoTagCollection
from supervisely.project.video_project import (
    VideoDataset,
    VideoProject,
    download_video_project,
    upload_video_project,
)
from supervisely.video import video

import supervisely.labeling_jobs.utils as lj

from supervisely.pointcloud import pointcloud
from supervisely.pointcloud_episodes import pointcloud_episodes
from supervisely.pointcloud_annotation.pointcloud_annotation import PointcloudAnnotation
from supervisely.pointcloud_annotation.pointcloud_episode_annotation import (
    PointcloudEpisodeAnnotation,
)
from supervisely.pointcloud_annotation.pointcloud_episode_frame import (
    PointcloudEpisodeFrame,
)
from supervisely.pointcloud_annotation.pointcloud_episode_frame_collection import (
    PointcloudEpisodeFrameCollection,
)
from supervisely.pointcloud_annotation.pointcloud_episode_object import (
    PointcloudEpisodeObject,
)
from supervisely.pointcloud_annotation.pointcloud_episode_object_collection import (
    PointcloudEpisodeObjectCollection,
)
from supervisely.pointcloud_annotation.pointcloud_episode_tag import (
    PointcloudEpisodeTag,
)
from supervisely.pointcloud_annotation.pointcloud_episode_tag_collection import (
    PointcloudEpisodeTagCollection,
)
from supervisely.pointcloud_annotation.pointcloud_object import PointcloudObject
from supervisely.pointcloud_annotation.pointcloud_figure import PointcloudFigure
from supervisely.pointcloud_annotation.pointcloud_tag import PointcloudTag
from supervisely.pointcloud_annotation.pointcloud_tag_collection import (
    PointcloudTagCollection,
)
from supervisely.project.pointcloud_project import (
    PointcloudDataset,
    PointcloudProject,
    download_pointcloud_project,
    upload_pointcloud_project,
)
from supervisely.project.pointcloud_episode_project import (
    PointcloudEpisodeDataset,
    PointcloudEpisodeProject,
    download_pointcloud_episode_project,
    upload_pointcloud_episode_project,
)

from supervisely.pyscripts_utils import utils as ps
from supervisely.io import docker_utils

import supervisely.app as app
from supervisely.app.fastapi import Application
from supervisely.app.v1.app_service import AppService
import supervisely.nn as nn

from supervisely.decorators.profile import timeit
from supervisely.decorators.profile import update_fields
from supervisely.decorators.inference import (
    process_image_roi,
    process_image_sliding_window,
)

import supervisely.script as script
from supervisely.user.user import UserRoleName
from supervisely.io import github_utils as git

from supervisely.aug import imgaug_utils

import supervisely.volume as volume
from supervisely.volume_annotation.volume_annotation import VolumeAnnotation
from supervisely.volume_annotation.volume_object import VolumeObject
from supervisely.volume_annotation.volume_object_collection import VolumeObjectCollection
from supervisely.volume_annotation.volume_tag import VolumeTag
from supervisely.volume_annotation.volume_tag_collection import VolumeTagCollection
from supervisely.volume_annotation.volume_figure import VolumeFigure
from supervisely.volume_annotation.slice import Slice
from supervisely.volume_annotation.plane import Plane
from supervisely.project.volume_project import (
    VolumeDataset,
    VolumeProject,
    download_volume_project,
    upload_volume_project,
)

from supervisely.convert.converter import ImportManager
from supervisely.convert.base_converter import AvailableImageConverters, BaseConverter

from supervisely.geometry.bitmap import SkeletonizeMethod

import supervisely.team_files as team_files
import supervisely.output as output

# start monkey patching
import importlib
import inspect
from supervisely.task.progress import tqdm_sly
import tqdm

from supervisely import convert

_original_tqdm = tqdm.tqdm


def get_module_names_from_stack(is_reversed=False) -> list:
    frame_records = inspect.stack()
    module_names = []
    for frame_record in frame_records:
        module_name = frame_record.frame.f_globals["__name__"]
        module_names.append(module_name)
    if is_reversed:
        module_names.reverse()
    return module_names


module_names = get_module_names_from_stack(is_reversed=True)

for mname in module_names:  # list starts with "__main__"
    m = importlib.import_module(mname)
    if not hasattr(m, "tqdm"):
        continue
    else:
        if hasattr(m.tqdm, "tqdm"):
            if type(m.tqdm.tqdm) != tqdm_sly:
                m.tqdm.tqdm = tqdm_sly
        else:
            if type(m.tqdm) != tqdm_sly:
                m.tqdm = tqdm_sly

# end monkeypatching

from supervisely.io.exception_handlers import handle_exceptions
from supervisely.app.fastapi.subapp import Event

try:
    setup_certificates()
except Exception as e:
    logger.warn(f"Failed to setup certificates. Reason: {repr(e)}", exc_info=True)

# Configure minimum instance version automatically from versions.json
# if new changes in Supervisely Python SDK require upgrade of the Supervisely instance.
# If the instance version is lower than the SDK version, users can face compatibility issues.
from supervisely.io.env import configure_minimum_instance_version

configure_minimum_instance_version()

LARGE_ENV_PLACEHOLDER = "@.@SLY_LARGE_ENV@.@"


def restore_env_vars():
    try:
        large_env_keys = []
        for key, value in os.environ.items():
            if value == LARGE_ENV_PLACEHOLDER:
                large_env_keys.append(key)
        if len(large_env_keys) == 0:
            return

        if utils.is_development():
            logger.info(
                "Large environment variables detected. Skipping restoration in development mode.",
                extra={"keys": large_env_keys},
            )
            return

        unknown_keys = []
        state_keys = []
        context_keys = []
        for key in large_env_keys:
            if key == "CONTEXT" or key.startswith("context."):
                context_keys.append(key)
            elif key.startswith("MODAL_STATE") or key.startswith("modal.state."):
                state_keys.append(key)
            else:
                unknown_keys.append(key)

        if state_keys or context_keys:
            api = Api()
            if state_keys:
                task_info = api.task.get_info_by_id(env.task_id())
                state = task_info.get("meta", {}).get("params", {}).get("state", {})
                modal_state_envs = json.flatten_json(state)
                modal_state_envs = json.modify_keys(modal_state_envs, prefix="modal.state.")

                restored_keys = []
                not_found_keys = []
                for key in state_keys:
                    if key == "MODAL_STATE":
                        os.environ[key] = json.json.dumps(state)
                    elif key in modal_state_envs:
                        os.environ[key] = str(modal_state_envs[key])
                    elif key.replace("_", ".") in [k.upper() for k in modal_state_envs]:
                        # some env vars do not support dots in their names
                        k = next(k for k in modal_state_envs if k.upper() == key.replace("_", "."))
                        os.environ[key] = str(modal_state_envs[k])
                    else:
                        not_found_keys.append(key)
                        continue
                    restored_keys.append(key)

                if restored_keys:
                    logger.info(
                        "Restored large environment variables from task state",
                        extra={"keys": restored_keys},
                    )

                if not_found_keys:
                    logger.warning(
                        "Failed to restore some large environment variables from task state. "
                        "No such keys in the state.",
                        extra={"keys": not_found_keys},
                    )

            if context_keys:
                context = api.task.get_context(env.task_id())
                context_envs = json.flatten_json(context)
                context_envs = json.modify_keys(context_envs, prefix="context.")

                restored_keys = []
                not_found_keys = []
                for key in context_keys:
                    if key == "CONTEXT":
                        os.environ[key] = json.json.dumps(context)
                    elif key in context_envs:
                        os.environ[key] = context_envs[key]
                    else:
                        not_found_keys.append(key)
                        continue
                    restored_keys.append(key)

                if restored_keys:
                    logger.info(
                        "Restored large environment variables from task context",
                        extra={"keys": restored_keys},
                    )

                if not_found_keys:
                    logger.warning(
                        "Failed to restore some large environment variables from task context. "
                        "No such keys in the context.",
                        extra={"keys": not_found_keys},
                    )

        if unknown_keys:
            logger.warning(
                "Found unknown large environment variables. Can't restore them.",
                extra={"keys": unknown_keys},
            )

    except Exception as e:
        logger.warning(
            "Failed to restore large environment variables.",
            exc_info=True,
        )


restore_env_vars()



"""API for training neural network models in Supervisely."""

from __future__ import annotations

from abc import abstractmethod
from typing import Any, Dict, List, Literal, Optional, Union

import supervisely.io.env as env
from supervisely import ProjectMeta
from supervisely._utils import logger
from supervisely.api.api import Api
from supervisely.api.nn.utils import (
    find_agent,
    find_apps_by_framework,
    find_team_by_path,
    get_artifacts_dir_and_checkpoint_name,
    get_experiment_info_by_task_id,
    run_train_app,
)
from supervisely.nn.experiments import get_experiment_info_by_artifacts_dir


class Model:
    """
    This class normalizes the user input into a structure that can be embedded into an
    application state (`guiState["model"]`) for a training app.

    The training UI supports two model sources:
    - **Pretrained models**: referenced by `"framework/model_name"` (e.g. `"RT-DETRv2/RT-DETRv2-M"`).
    - **Custom models**: referenced by a path to a checkpoint in Team Files (starts with `"/"`),
      or by a path without leading slash if it exists in Team Files.
    """

    def __init__(
        self,
        api: "Api",
        source: str,
        framework: str = None,
        model_name: str = None,
        team_id: int = None,
        artifacts_dir: str = None,
        checkpoint_name: str = None
    ):
        """
        Initialize a :class:`~supervisely.api.nn.train_api.Model` instance.

        :param api: Supervisely API client.
        :type api: :class:`~supervisely.api.api.Api`
        :param source: Source of the model.
        :type source: str
        :param framework: Framework of the model.
        :type framework: str
        :param model_name: Name of the model.
        :type model_name: str
        :param team_id: Team id of the model.
        :type team_id: int
        :param artifacts_dir: Artifacts directory of the model.
        :type artifacts_dir: str
        :param checkpoint_name: Checkpoint name of the model.
        :type checkpoint_name: str
        """
        self.api = api
        self.source: str = source
        self.framework: str = framework
        self.model_name: str = model_name
        self.team_id: int = team_id
        self.artifacts_dir: str = artifacts_dir
        self.checkpoint_name: str = checkpoint_name
        self._init()

    def _init(self):
        """
        Finalize initialization for custom models.

        For custom checkpoints, we resolve experiment metadata to determine:
        - `task_id` of the experiment that produced artifacts.
        - `framework` name used by that experiment.
        """
        if self.source == "Custom models":
            if self.team_id is None:
                self.team_id = find_team_by_path(self.api, self.artifacts_dir, team_id=self.team_id)
            experiment_info = get_experiment_info_by_artifacts_dir(self.api, self.team_id, self.artifacts_dir)
            if not experiment_info:
                raise ValueError(
                    f"Failed to retrieve experiment info for artifacts_dir: '{self.artifacts_dir}'"
                )
            self.task_id = experiment_info.task_id
            self.framework = experiment_info.framework_name

    @classmethod
    def parse(cls, api: "Api", model: str) -> Model:
        """
        Parse a user-provided model identifier into a `Model`.

        :param api: Supervisely API client.
        :type api: :class:`~supervisely.api.api.Api`
        :param model: Either a checkpoint path starting with "/" (Team Files), a checkpoint path
            without leading slash (will be checked in Team Files), or a pretrained model name in
            format "framework/model_name" (e.g. "RT-DETRv2/RT-DETRv2-M", "YOLO/YOLO26s-det").
        :type model: str
        :returns: Parsed model reference ready to be embedded into training app state.
        :rtype: :class:`~supervisely.api.nn.train_api.Model`
        :raises ValueError: If the string cannot be parsed or required metadata cannot be resolved.
        """
        checkpoint = None
        pretrained = model
        team_id = None
        if model.startswith("/"):
            checkpoint = model
        else:
            found_team_id = find_team_by_path(api,
                f"/{model}", team_id=team_id, raise_not_found=False
            )
            if found_team_id is not None:
                checkpoint = f"/{model}"
                team_id = found_team_id
                logger.debug(f"Found checkpoint in team {team_id}")
            else:
                pretrained = model

        if checkpoint is not None:
            artifacts_dir, checkpoint_name = get_artifacts_dir_and_checkpoint_name(checkpoint)
            return cls(
                api=api,
                source="Custom models",
                artifacts_dir=artifacts_dir,
                checkpoint_name=checkpoint_name,
                team_id=team_id,
            )
        else:
            framework, model_name = pretrained.split("/", 1)
            return cls(
                api=api, source="Pretrained models", framework=framework, model_name=model_name
            )

    def app_state(self):
        """Return a JSON-serializable dict for `guiState["model"]`."""
        if self.source == "Pretrained models":
            return {
                "source": self.source,
                "model_name": self.model_name,
            }
        else:
            return {
                "source": self.source,
                "task_id": self.task_id,
                "checkpoint": self.checkpoint_name
            }


class _TrainValSplit:
    """Base class for Train/Val split strategies used by training UI."""

    @abstractmethod
    def app_state(self) -> Dict[str, Any]:
        """Return a JSON-serializable dict for `guiState["train_val_split"]`.

        :returns: Dict for training app UI state.
        :rtype: Dict[str, Any]
        """
        raise NotImplementedError()


class RandomSplit(_TrainValSplit):
    """Split dataset randomly by percent.

    :param percent: Percent of the dataset to split into train.
    :type percent: int
    :param split: Split method: "train" or "val".
    :type split: str
    """

    def __init__(self, percent: int = 80, split: str = "train"):
        self.percent = percent
        self.split = split

    def app_state(self):
        """Return a JSON-serializable dict for `guiState["train_val_split"]`."""
        return {
            "method": "random",
            "split": self.split,
            "percent": self.percent
        }

class DatasetsSplit(_TrainValSplit):
    """Split by explicit train/val dataset ids inside the project.

    :param train_datasets: List of dataset ids to split into train.
    :type train_datasets: List[int]
    :param val_datasets: List of dataset ids to split into val.
    :type val_datasets: List[int]
    """

    def __init__(self, train_datasets: List[int], val_datasets: List[int]):
        self.train_datasets = train_datasets
        self.val_datasets = val_datasets

    def app_state(self):
        """Serialize split settings for training UI."""
        return {
            "method": "datasets",
            "train_datasets": self.train_datasets,
            "val_datasets": self.val_datasets
        }

class TagsSplit(_TrainValSplit):
    """
    Split by tags: items with `train_tag` go to train, `val_tag` to val.

    :param train_tag: Tag to split into train.
    :type train_tag: str
    :param val_tag: Tag to split into val.
    :type val_tag: str
    :param untagged_action: Action to take for untagged items: "train", "val", "ignore".
    :type untagged_action: Literal["train", "val", "ignore"]
    """

    def __init__(self, train_tag: str, val_tag: str, untagged_action: Literal["train", "val", "ignore"]):
        self.train_tag = train_tag
        self.val_tag = val_tag
        self.untagged_action = untagged_action

    def app_state(self):
        """Serialize split settings for training UI."""
        return {
            "method": "tags",
            "train_tag": self.train_tag,
            "val_tag": self.val_tag,
            "untagged_action": self.untagged_action
        }

class CollectionsSplit(_TrainValSplit):
    """
    Split by entity collections (train collections vs val collections).

    :param train_collections: List of collection ids to split into train.
    :type train_collections: List[int]
    :param val_collections: List of collection ids to split into val.
    :type val_collections: List[int]
    """

    def __init__(self, train_collections: List[int], val_collections: List[int]):
        self.train_collections = train_collections
        self.val_collections = val_collections

    def app_state(self):
        """Serialize split settings for training UI."""
        return {
            "method": "collections",
            "train_collections": self.train_collections,
            "val_collections": self.val_collections
        }

class TrainApi:
    """High-level API to start a training application.

    You can read more about the training API in the [Training API documentation](https://developer.supervisely.com/advanced-user-guide/automate-with-python-sdk-and-api/training-api).

    This wrapper prepares the `params`/`state` payload expected by the training UI app
    and starts an app task on a given agent.

    Typical usage:

    - Choose a model (pretrained or custom checkpoint)
    - Provide training settings (model, classes, train/val split, hyperparameters, etc.)
    - Start the training app
    """

    def __init__(self, api: "Api"):
        """
        Create a :class:`~supervisely.api.nn.train_api.TrainApi` instance.

        :param api: Supervisely API client.
        :type api: :class:`~supervisely.api.api.Api`
        """
        self._api = api

    def _get_app_state(
        self,
        project_id: int,
        model: Model,
        classes: List[str],
        train_val_split: _TrainValSplit,
        experiment_name: str = None,
        hyperparameters: str = None,
        convert_class_shapes: bool = True,
        enable_benchmark: bool = True,
        enable_speedtest: bool = False,
        cache_project: bool = True,
        export_onnx: bool = False,
        export_tensorrt: bool = False,
        autostart: bool = True,
    ):
        """
        Build training app state payload.

        The resulting structure is passed to the training app as `params` (task arguments).
        It follows the `TrainApp`/GUI expected schema:

        - `state.slyProjectId`: project id
        - `state.guiState`: UI state (model/classes/split/hyperparameters/options/experiment_name/start_training)

        Notes:
            - `hyperparameters` is expected to be a YAML string (or `None` to keep defaults from training app).

        :param project_id: Project id to train on.
        :type project_id: int
        :param model: Parsed model reference.
        :type model: Model
        :param classes: Class names to train on (filtered to project meta upstream).
        :type classes: List[str]
        :param train_val_split: Train/Val split strategy for the GUI.
        :type train_val_split: :class:`~supervisely.api.nn.train_api._TrainValSplit`
        :param experiment_name: Optional experiment name shown in UI and used for artifacts.
        :type experiment_name: str, optional
        :param hyperparameters: Hyperparameters YAML string for the training app. If None, GUI keeps defaults.
        :type hyperparameters: str, optional
        :param convert_class_shapes: Whether to auto-convert shapes to framework requirements.
        :type convert_class_shapes: bool
        :param enable_benchmark: Enable post-training evaluation (Model Benchmark).
        :type enable_benchmark: bool
        :param enable_speedtest: Enable speed test as part of benchmark.
        :type enable_speedtest: bool
        :param cache_project: Cache project on agent before training.
        :type cache_project: bool
        :param export_onnx: Enable export to ONNXRuntime (if supported by the app/framework).
        :type export_onnx: bool
        :param export_tensorrt: Enable export to TensorRT engine (if supported by the app/framework).
        :type export_tensorrt: bool
        :param autostart: If True, training is started automatically after UI state is applied.
        :type autostart: bool
        :returns: Task params payload for the training app.
        :rtype: Dict[str, Any]
        """
        app_state = {
            "state": {
                # 1. Project
                "slyProjectId": project_id,
                "guiState": {
                    # 2. Model
                    "model": model.app_state(),
                    # 3. Classes
                    "classes": classes,
                    # 4. Train/Val Split
                    "train_val_split": train_val_split.app_state(),
                    # 5. Hyperparameters
                    "hyperparameters": hyperparameters,  # yaml string
                    # 6. Options
                    "options": {
                        "convert_class_shapes": convert_class_shapes,
                        "model_benchmark": {
                            "enable": enable_benchmark or enable_speedtest,
                            "speed_test": enable_speedtest,
                        },
                        "cache_project": cache_project,
                        "export": {
                            "enable": export_onnx or export_tensorrt,
                            "ONNXRuntime": export_onnx,
                            "TensorRT": export_tensorrt,
                        },
                    },
                    # 7. Experiment Name
                    "experiment_name": experiment_name,
                    # 8. Start Training
                    "start_training": autostart,  # Starts training automatically
                },
            },
        }
        return app_state

    def run(
        self,
        project_id: int,
        model: str,
        hyperparameters: str = None,
        experiment_name: str = None,
        classes: List[str] = None,
        train_val_split: Union[RandomSplit, DatasetsSplit, TagsSplit, CollectionsSplit] = None,
        convert_class_shapes: bool = True,
        enable_benchmark: bool = True,
        enable_speedtest: bool = False,
        cache_project: bool = True,
        export_onnx: bool = False,
        export_tensorrt: bool = False,
        autostart: bool = True,
        agent_id: int = None,
        **kwargs,
    ):
        """
        Start a training application task for a project.

        You can read more about the training API in the [Training API documentation](https://developer.supervisely.com/advanced-user-guide/automate-with-python-sdk-and-api/training-api).

        :param agent_id: Agent ID where the app task will run.
        :type agent_id: int
        :param project_id: Project ID to train on.
        :type project_id: int
        :param model: Either a checkpoint path in Team Files (e.g. "/experiments/.../checkpoints/best.pth"),
            or a pretrained model name in format "framework/model_name" (e.g. "RT-DETRv2/RT-DETRv2-M", "YOLO/YOLO26s-det").
        :type model: str
        :param hyperparameters: Hyperparameters YAML string for the training app. If None, uses defaults from training app.
        :type hyperparameters: str, optional
        :param experiment_name: Optional experiment name used in training app. Will be auto-generated if not provided.
        :type experiment_name: str, optional
        :param classes: Optional subset of class names to train on. If provided, names not present in project meta are ignored.
        :type classes: List[str], optional
        :param train_val_split: Optional split strategy; defaults to :class:`~supervisely.api.nn.train_api.RandomSplit`.
        :type train_val_split: Union[:class:`~supervisely.api.nn.train_api.RandomSplit`, :class:`~supervisely.api.nn.train_api.DatasetsSplit`, :class:`~supervisely.api.nn.train_api.TagsSplit`, :class:`~supervisely.api.nn.train_api.CollectionsSplit`], optional
        :param convert_class_shapes: Whether to convert class shapes to framework requirements automatically.
        :type convert_class_shapes: bool, optional
        :param enable_benchmark: Enable post-training evaluation (Model Benchmark) after training.
        :type enable_benchmark: bool, optional
        :param enable_speedtest: Enable speed test as part of benchmark after training.
        :type enable_speedtest: bool, optional
        :param cache_project: Cache project on agent before training to save time on downloading project next time.
        :type cache_project: bool, optional
        :param export_onnx: Enable export to ONNXRuntime (if supported by the training app/framework).
        :type export_onnx: bool, optional
        :param export_tensorrt: Enable export to TensorRT engine (if supported by the training app/framework).
        :type export_tensorrt: bool, optional
        :param autostart: If True, training is started automatically after all settings are applied. If False, training must be started manually from the training app UI by clicking the "Start Training" button.
        :type autostart: bool, optional
        :returns: Task information dict for the created app task.
        :rtype: Dict[str, Any]
        :raises ValueError: If a suitable training app cannot be found for the detected framework.

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly
                from supervisely.api.nn.train_api import TrainApi

                if sly.is_development():
                    load_dotenv("local.env")
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                agent_id = sly.env.agent_id()
                project_id = sly.env.project_id()

                train = TrainApi(api)
                train.run(agent_id, project_id, model="YOLO/YOLO11n-det")
        """
        model: Model = Model.parse(self._api, model)
        return self._run(
            project_id=project_id,
            model=model,
            hyperparameters=hyperparameters,
            experiment_name=experiment_name,
            classes=classes,
            train_val_split=train_val_split,
            convert_class_shapes=convert_class_shapes,
            enable_benchmark=enable_benchmark,
            enable_speedtest=enable_speedtest,
            cache_project=cache_project,
            export_onnx=export_onnx,
            export_tensorrt=export_tensorrt,
            autostart=autostart,
            agent_id=agent_id,
            **kwargs
        )

    def _run(
        self,
        project_id: int,
        model: Model,
        hyperparameters: str = None,
        experiment_name: str = None,
        classes: List[str] = None,
        train_val_split: Union[RandomSplit, DatasetsSplit, TagsSplit, CollectionsSplit] = None,
        convert_class_shapes: bool = True,
        enable_benchmark: bool = True,
        enable_speedtest: bool = True,
        cache_project: bool = True,
        export_onnx: bool = False,
        export_tensorrt: bool = False,
        autostart: bool = True,
        agent_id: int = None,
        **kwargs
    ) -> Dict:
        project_info = self._api.project.get_info_by_id(project_id)
        workspace_id = project_info.workspace_id
        team_id = project_info.team_id

        project_meta_json = self._api.project.get_meta(project_id)
        project_meta = ProjectMeta.from_json(project_meta_json)
        if classes:
            classes = [obj_class.name for obj_class in project_meta.obj_classes if obj_class.name in classes]
        else:
            classes = [obj_class.name for obj_class in project_meta.obj_classes]

        if train_val_split is None:
            train_val_split = RandomSplit()

        module = self.find_train_app_by_framework(model.framework)
        if module is None:
            raise ValueError(f"Failed to detect train app by framework: '{model.framework}'")
        module_id = module["id"]

        if agent_id is None:
            agent_id = find_agent(self._api, team_id)

        app_state = self._get_app_state(
            project_id=project_id,
            model=model,
            classes=classes,
            train_val_split=train_val_split,
            experiment_name=experiment_name,
            hyperparameters=hyperparameters,
            convert_class_shapes = convert_class_shapes,
            enable_benchmark = enable_benchmark,
            enable_speedtest = enable_speedtest,
            cache_project = cache_project,
            export_onnx = export_onnx,
            export_tensorrt = export_tensorrt,
            autostart = autostart,
        )
        task_info = run_train_app(
            api=self._api,
            agent_id=agent_id,
            module_id=module_id,
            workspace_id=workspace_id,
            app_state=app_state,
            timeout=100,
            **kwargs
        )
        return task_info

    def find_train_app_by_framework(self, framework: str):
        """
        Find a training app module for the given framework.

        :param framework: Framework name (e.g. "RT-DETRv2", "YOLO", "DEIM").
        :type framework: str
        :returns: Ecosystem module dict (as returned by the API) or None if not found.
        :rtype: Union[dict, None]
        """
        modules = find_apps_by_framework(self._api, framework, ["train"])
        if not modules:
            return None
        return modules[0]

    def finetune(
        self,
        task_id: int,
        project_id: int = None,
        agent_id: int = None,
        **kwargs
    ) -> Dict:
        """
        Finetune best checkpoint from a previous training task.

        :param task_id: Task ID to finetune best checkpoint from.
        :type task_id: int
        :param project_id: Project ID to train on. If not provided, will use project ID from the provided Task ID
        :type project_id: int
        :param agent_id: Agent ID where the app task will run. If not provided, will use the most available agent from the current team.
        :type agent_id: int
        :returns: Task information dict for the created app task.
        :rtype: Dict[str, Any]
        :raises ValueError: If a suitable training app cannot be found for the detected framework.

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly
                from supervisely.api.nn.train_api import TrainApi

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv("local.env")
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                agent_id = sly.env.agent_id()
                project_id = sly.env.project_id()

                train = TrainApi(api)
                train.finetune(agent_id, project_id, task_id=123)
        """
        experiment_info = get_experiment_info_by_task_id(self._api, task_id)
        if experiment_info is None:
            raise ValueError(f"Not found experiment data for task {task_id}")
        module = self.find_train_app_by_framework(experiment_info.framework_name)
        if module is None:
            raise ValueError(f"Failed to detect train app by framework: '{experiment_info.framework_name}'")
        if project_id is None:
            project_id = experiment_info.project_id
        project_info = self._api.project.get_info_by_id(project_id)
        team_id = project_info.team_id
        if agent_id is None:
            agent_id = find_agent(self._api, team_id)
        module_id = module["id"]
        app_state = {"state": {"trainTaskId": task_id, "trainMode": "continue", "slyProjectId": project_id}}
        task_info = run_train_app(
            api=self._api,
            agent_id=agent_id,
            module_id=module_id,
            timeout=100,
            workspace_id=project_info.workspace_id,
            app_state=app_state,
            **kwargs
        )
        return task_info

    def rerun_from_task_id(
        self, task_id: int, project_id: int = None, agent_id: int = None, **kwargs
    ):
        """
        Rerun a training application task from a previous task.

        :param task_id: Task ID to rerun from.
        :type task_id: int
        :param project_id: Project ID to train on. If not provided, will use project ID from the provided Task ID
        :type project_id: int
        :param agent_id: Agent ID where the app task will run. If not provided, will use the most available agent from the current team.
        :type agent_id: int
        :returns: Task information dict for the created app task.
        :rtype: Dict[str, Any]
        :raises ValueError: If a suitable training app cannot be found for the detected framework.

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly
                from supervisely.api.nn.train_api import TrainApi

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv("local.env")
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                agent_id = sly.env.agent_id()
                project_id = sly.env.project_id()

                train = TrainApi(api)
                train.rerun_from_task_id(agent_id, project_id, task_id=123)
        """
        experiment_info = get_experiment_info_by_task_id(self._api, task_id)
        if experiment_info is None:
            raise ValueError(f"Not found experiment data for task {task_id}")
        module = self.find_train_app_by_framework(experiment_info.framework_name)
        if module is None:
            raise ValueError(f"Failed to detect train app by framework: '{experiment_info.framework_name}'")
        if project_id is None:
            project_id = experiment_info.project_id
        project_info = self._api.project.get_info_by_id(project_id)
        team_id = project_info.team_id
        if agent_id is None:
            agent_id = find_agent(self._api, team_id)
        module_id = module["id"]
        app_state = {"state": {"trainTaskId": task_id, "trainMode": "new", "slyProjectId": project_id}}
        task_info = run_train_app(
            api=self._api,
            agent_id=agent_id,
            module_id=module_id,
            timeout=100,
            workspace_id=project_info.workspace_id,
            app_state=app_state,
            **kwargs
        )
        return task_info



# coding: utf-8

# docs
from __future__ import annotations

from copy import deepcopy
from math import ceil, floor
from typing import Dict, List, Optional, Tuple, Union

import cv2
import numpy as np

from supervisely.geometry.constants import (
    ANGLE,
    CLASS_ID,
    CREATED_AT,
    ID,
    LABELER_LOGIN,
    POINTS,
    UPDATED_AT,
)
from supervisely.geometry.geometry import Geometry
from supervisely.geometry.point_location import PointLocation, points_to_row_col_list
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.rectangle import Rectangle


class OrientedBBox(Rectangle):
    """
    OrientedBBox geometry for a single :class:`Label<supervisely.annotation.label.Label>`. :class:`OrientedBBox<OrientedBBox>` class object is immutable.

    :param top: Minimal vertical value of OrientedBBox object.
    :type top: int or float
    :param left: Minimal horizontal value of OrientedBBox object.
    :type left: int or float
    :param bottom: Maximal vertical value of OrientedBBox object.
    :type bottom: int or float
    :param right: Maximal vertical value of OrientedBBox object.
    :type right: int or float
    :param angle: Angle of rotation in radians. Positive values mean clockwise rotation.
    :type angle: int or float, optional
    :param sly_id: OrientedBBox ID in Supervisely server.
    :type sly_id: int, optional
    :param class_id: ID of :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` to which OrientedBBox belongs.
    :type class_id: int, optional
    :param labeler_login: Login of the user who created OrientedBBox.
    :type labeler_login: str, optional
    :param updated_at: Date and Time when OrientedBBox was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
    :type updated_at: str, optional
    :param created_at: Date and Time when OrientedBBox was created. Date Format is the same as in "updated_at" parameter.
    :type created_at: str, optional
    :raises: :class:`ValueError`. OrientedBBox top argument must have less or equal value then bottom, left argument must have less or equal value then right

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        import math

        top = 100
        left = 100
        bottom = 700
        right = 900
        angle = math.pi / 12  # 15 degrees in radians
        figure = sly.OrientedBBox(top, left, bottom, right, angle=angle)
    """

    @staticmethod
    def geometry_name():
        """ """
        return "oriented_bbox"

    def __init__(
        self,
        top: int,
        left: int,
        bottom: int,
        right: int,
        angle: Union[int, float] = 0,
        sly_id: Optional[int] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[str] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
    ):

        super().__init__(
            top=top,
            left=left,
            bottom=bottom,
            right=right,
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )
        self._angle = angle
        if self._angle is None:
            self._angle = 0

    @property
    def angle(self) -> Union[int, float]:
        """
        Angle of rotation in radians. Positive values mean clockwise rotation.

        :return: Angle of rotation in radians
        :rtype: int or float
        :Usage example:

         .. code-block:: python

            angle = oriented_bbox.angle
        """
        return self._angle

    def to_json(self) -> Dict:
        """
        Convert the OrientedBBox to a json dict. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :return: Json format as a dict
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            figure_json = figure.to_json()
            print(figure_json)
            # Output: {
            #    "points": [
            #        [100, 100],
            #        [900, 700]
            #    ],
            #    "angle": 0.2618  # radians (15 degrees)
            # }
        """
        packed_obj = {
            POINTS: points_to_row_col_list(self._points, flip_row_col_order=True),
            ANGLE: self._angle,
        }
        self._add_creation_info(packed_obj)
        return packed_obj

    @classmethod
    def from_json(cls, data: Dict) -> OrientedBBox:
        """
        Convert a json dict to OrientedBBox. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: OrientedBBox in json format as a dict.
        :type data: dict
        :return: OrientedBBox object
        :rtype: :class:`OrientedBBox<OrientedBBox>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            import math

            figure_json = {
                "points": [
                    [100, 100],
                    [900, 700]
                ],
                "angle": math.pi / 12  # 15 degrees in radians
            }
            figure = sly.OrientedBBox.from_json(figure_json)
        """
        if POINTS not in data:
            raise ValueError("Input data must contain {} field.".format(POINTS))
        if ANGLE not in data:
            raise ValueError("Input data must contain {} field.".format(ANGLE))
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        angle = data.get(ANGLE, 0)

        exterior = data[POINTS]
        if len(exterior) != 2:
            raise ValueError(
                '"exterior" field must contain exactly two points to create OrientedBBox object.'
            )
        [top, bottom] = sorted([exterior[0][1], exterior[1][1]])
        [left, right] = sorted([exterior[0][0], exterior[1][0]])

        return cls(
            top=top,
            left=left,
            bottom=bottom,
            right=right,
            angle=angle,
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

    def to_bbox(self) -> Rectangle:
        """
        Convert the OrientedBBox to the axis-aligned :class:`Rectangle<supervisely.geometry.rectangle.Rectangle>` that fully contains the OrientedBBox.

        :return: Axis-aligned Rectangle object
        :rtype: :class:`Rectangle<supervisely.geometry.rectangle.Rectangle>`
        :Usage example:
            .. code-block:: python
    
                axis_aligned_bbox = oriented_bbox.to_bbox()
            """
        two_pi = 2 * np.pi
        if self._angle % two_pi == 0:
            return Rectangle(
                top=self.top,
                left=self.left,
                bottom=self.bottom,
                right=self.right,
                sly_id=self.sly_id,
                class_id=self.class_id,
                labeler_login=self.labeler_login,
                updated_at=self.updated_at,
                created_at=self.created_at,
            )

        cos_angle = abs(np.cos(self._angle))
        sin_angle = abs(np.sin(self._angle))

        new_w = self.width * cos_angle + self.height * sin_angle
        new_h = self.width * sin_angle + self.height * cos_angle

        new_left = self.center.col - new_w / 2.0
        new_right = self.center.col + new_w / 2.0
        new_top = self.center.row - new_h / 2.0
        new_bottom = self.center.row + new_h / 2.0

        return Rectangle(
            top=new_top,
            left=new_left,
            bottom=new_bottom,
            right=new_right,
            sly_id=self.sly_id,
            class_id=self.class_id,
            labeler_login=self.labeler_login,
            updated_at=self.updated_at,
            created_at=self.created_at,
        )

    def _transform(self, transform_fn):
        """ """
        transformed_corners = [transform_fn(p) for p in self.corners]
        rows, cols = zip(*points_to_row_col_list(transformed_corners))
        return OrientedBBox(
            top=round(min(rows)),
            left=round(min(cols)),
            bottom=round(max(rows)),
            right=round(max(cols)),
            angle=self._angle,
        )

    def contains_point_location(self, point: PointLocation) -> bool:
        """
        Check if the OrientedBBox contains the given point.

        :param point: PointLocation object
        :type point: :class:`PointLocation<supervisely.geometry.point_location.PointLocation>`
        :return: True if the point is inside the OrientedBBox, False otherwise
        :rtype: bool
        :Usage example:

         .. code-block:: python

            point = sly.PointLocation(150, 200)
            is_inside = oriented_bbox.contains_point_location(point)
        """
        # Rotate point in the opposite direction around the center of the oriented bbox
        cos_angle = np.cos(-self._angle)
        sin_angle = np.sin(-self._angle)

        # Translate point to origin
        translated_x = point.col - self.center.col
        translated_y = point.row - self.center.row

        # Rotate point
        rotated_x = translated_x * cos_angle - translated_y * sin_angle
        rotated_y = translated_x * sin_angle + translated_y * cos_angle

        # Translate point back
        final_x = rotated_x + self.center.col
        final_y = rotated_y + self.center.row

        # Check if the rotated point is within the axis-aligned bbox
        return self.left <= final_x <= self.right and self.top <= final_y <= self.bottom

    def contains_point(self, geometry: Geometry) -> bool:
        """
        Check if the OrientedBBox contains the given point.

        :param geometry: PointLocation object
        :type geometry: :class:`Geometry<supervisely.geometry.geometry.Geometry>`
        :return: True if the point is inside the OrientedBBox, False otherwise
        :rtype: bool
        :Usage example:

         .. code-block:: python

            point = sly.PointLocation(150, 200)
            is_inside = oriented_bbox.contains_point(point)
        """
        if isinstance(geometry, PointLocation):
            return self.contains_point_location(geometry)
        elif isinstance(geometry, Rectangle):
            return self.contains_rectangle(geometry)
        elif isinstance(geometry, OrientedBBox):
            return self.contains_obb(geometry)
        else:
            raise TypeError(
                "Unsupported geometry type for contains_point method. "
                "Supported types are PointLocation, Rectangle, and OrientedBBox."
            )

    def contains_obb(self, obb: OrientedBBox) -> bool:
        """
        Check if the OrientedBBox contains the given OrientedBBox.

        :param obb: OrientedBBox object
        :type obb: :class:`OrientedBBox<supervisely.geometry.oriented_bbox.OrientedBBox>`
        :return: True if the OrientedBBox is inside the OrientedBBox, False otherwise
        :rtype: bool
        :Usage example:

         .. code-block:: python

            obb2 = sly.OrientedBBox(150, 200, 400, 500, angle=10)
            is_inside = obb1.contains_obb(obb2)
        """
        # Get the corners of the obb
        corners = obb.calculate_rotated_corners()

        # Check if all corners are inside the current obb
        for corner in corners:
            if not self.contains_point_location(corner):
                return False
        return True

    @staticmethod
    def _calculate_rotated_corners(obb: OrientedBBox) -> List[PointLocation]:
        """
        Get the corners of the OrientedBBox.

        :return: List of corners as (x, y) tuples
        :rtype: List[Tuple[float, float]]
        :Usage example:

         .. code-block:: python

            corners = oriented_bbox.calculate_rotated_corners()
        """
        cos_angle = np.cos(obb._angle)
        sin_angle = np.sin(obb._angle)

        rotated_corners = []
        for corner in obb.corners:  # [Top-left, Top-right, Bottom-right, Bottom-left]
            # First translate to origin (subtract center)
            translated_x = corner.col - obb.center.col
            translated_y = corner.row - obb.center.row

            # Then rotate
            rotated_x = translated_x * cos_angle - translated_y * sin_angle
            rotated_y = translated_x * sin_angle + translated_y * cos_angle

            # Then translate back (add center)
            final_x = rotated_x + obb.center.col
            final_y = rotated_y + obb.center.row

            rotated_corners.append(PointLocation(row=final_y, col=final_x))

        return rotated_corners

    def calculate_rotated_corners(self) -> List[PointLocation]:
        """
        Get the corners of the OrientedBBox.

        :return: List of corners as PointLocation objects
        :rtype: List[:class:`PointLocation<supervisely.geometry.point_location.PointLocation>`]
        :Usage example:

         .. code-block:: python

            corners = oriented_bbox.calculate_rotated_corners()
        """
        return self._calculate_rotated_corners(self)

    def contains_rectangle(self, rectangle: Rectangle) -> bool:
        """
        Check if the OrientedBBox contains the given Rectangle.

        :param rectangle: Rectangle object
        :type rectangle: :class:`Rectangle<supervisely.geometry.rectangle.Rectangle>`
        :return: True if the Rectangle is inside the OrientedBBox, False otherwise
        :rtype: bool
        :Usage example:

         .. code-block:: python

            rectangle = sly.Rectangle(150, 200, 400, 500)
            is_inside = obb.contains_rectangle(rectangle)
        """
        # Get the corners of the rectangle
        corners = rectangle.corners  # [Top-left, Top-right, Bottom-right, Bottom-left]

        # Check if all corners are inside the current obb
        for corner in corners:
            if not self.contains_point_location(corner):
                return False
        return True

    @classmethod
    def from_bbox(cls, bbox: Rectangle) -> OrientedBBox:
        """
        Create OrientedBBox from given Rectangle.

        :param bbox: Rectangle object.
        :type bbox: :class:`Rectangle<supervisely.geometry.rectangle.Rectangle>`
        :return: OrientedBBox object
        :rtype: :class:`OrientedBBox<OrientedBBox>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            axis_aligned_bbox = sly.Rectangle(100, 100, 700, 900)
            figure_from_bbox = sly.OrientedBBox.from_bbox(axis_aligned_bbox)
        """
        return cls(
            top=bbox.top,
            left=bbox.left,
            bottom=bbox.bottom,
            right=bbox.right,
            angle=0,
        )

    @classmethod
    def from_array(cls, arr: np.ndarray, angle: Union[int, float] = 0) -> OrientedBBox:
        """
        Create OrientedBBox with given array shape.

        :param arr: Numpy array.
        :type arr: np.ndarray
        :return: OrientedBBox object
        :rtype: :class:`OrientedBBox<OrientedBBox>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            np_array = np.zeros((300, 400))
            figure_from_np = sly.OrientedBBox.from_array(np_array, angle=15)
        """
        return cls(top=0, left=0, bottom=arr.shape[0] - 1, right=arr.shape[1] - 1, angle=angle)

    def get_cropped_numpy_slice(self, data: np.ndarray) -> np.ndarray:
        """
        Slice of given numpy array with OrientedBBox align bbox.

        :param data: Numpy array.
        :type data: np.ndarray
        :return: Sliced numpy array
        :rtype: :class:`np.ndarray<np.ndarray>`

        :Usage Example:

         .. code-block:: python

            np_slice = np.zeros((200, 500))
            mask_slice = figure.get_cropped_numpy_slice(np_slice)
            print(mask_slice.shape)
        """
        axis_aligned_bbox = self.to_bbox()
        top = max(0, floor(axis_aligned_bbox.top))
        left = max(0, floor(axis_aligned_bbox.left))
        bottom = min(data.shape[0], ceil(axis_aligned_bbox.bottom))
        right = min(data.shape[1], ceil(axis_aligned_bbox.right))
        return data[top:bottom, left:right]

    @classmethod
    def allowed_transforms(cls):
        """ """
        from supervisely.geometry.alpha_mask import AlphaMask
        from supervisely.geometry.any_geometry import AnyGeometry
        from supervisely.geometry.bitmap import Bitmap

        return [AlphaMask, AnyGeometry, Bitmap, Polygon, OrientedBBox]

    def crop(self, clip: OrientedBBox | Rectangle) -> List[OrientedBBox]:
        """Crop the OrientedBBox by another OrientedBBox using the Sutherland-Hodgman algorithm."""
        subject_corners = self._calculate_rotated_corners(self)
        if isinstance(clip, Rectangle):
            if all([clip.contains_point_location(corner) for corner in subject_corners]):
                return [self]
            clip_corners = clip.corners
        else:
            if clip.contains_obb(self):
                return [self]
            clip_corners = self._calculate_rotated_corners(clip)

        def inside(p: PointLocation, edge_start: PointLocation, edge_end: PointLocation) -> bool:
            cx1, cy1 = edge_start.col, edge_start.row
            cx2, cy2 = edge_end.col, edge_end.row
            px, py = p.col, p.row
            return (cx2 - cx1) * (py - cy1) > (cy2 - cy1) * (px - cx1)

        def compute_intersection(
            p1: PointLocation,
            p2: PointLocation,
            edge_start: PointLocation,
            edge_end: PointLocation,
        ) -> Optional[PointLocation]:
            x1, y1 = p1.col, p1.row
            x2, y2 = p2.col, p2.row
            x3, y3 = edge_start.col, edge_start.row
            x4, y4 = edge_end.col, edge_end.row
            denom = (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4)
            if abs(denom) < 1e-10: return None
            t = ((x1 - x3) * (y3 - y4) - (y1 - y3) * (x3 - x4)) / denom
            return PointLocation(row=y1 + t * (y2 - y1), col=x1 + t * (x2 - x1))

        output = subject_corners[:]
        n = len(clip_corners)

        for i in range(n):
            edge_start = clip_corners[i]
            edge_end = clip_corners[(i + 1) % n]
            input_list = output
            output = []

            m = len(input_list)
            for j in range(m):
                curr = input_list[j]
                prev = input_list[(j - 1) % m]

                curr_inside = inside(curr, edge_start, edge_end)
                prev_inside = inside(prev, edge_start, edge_end)

                if curr_inside:
                    if not prev_inside:
                        inter = compute_intersection(prev, curr, edge_start, edge_end)
                        if inter: output.append(inter)
                    output.append(curr)
                elif prev_inside:
                    inter = compute_intersection(prev, curr, edge_start, edge_end)
                    if inter: output.append(inter)

            if len(output) < 3: return []

        polygon = Polygon(output, [])
        bbox = polygon.to_bbox()
        return [OrientedBBox.from_bbox(bbox)]

    def _draw_contour_impl(self, bitmap, color, thickness=1, config=None):
        """ """
        corners = self.calculate_rotated_corners()
        pts = np.array([[int(corner.col), int(corner.row)] for corner in corners], np.int32)
        pts = pts.reshape((-1, 1, 2))
        cv2.polylines(bitmap, [pts], isClosed=True, color=color, thickness=thickness)

    def _draw_impl(self, bitmap: np.ndarray, color, thickness=1, config=None):
        """ """
        corners = self.calculate_rotated_corners()
        pts = np.array([[int(corner.col), int(corner.row)] for corner in corners], np.int32)
        pts = pts.reshape((-1, 1, 2))
        cv2.fillPoly(bitmap, [pts], color)

    @classmethod
    def _to_pixel_coordinate_system_json(cls, data: Dict, image_size: List[int]) -> Dict:
        """
        Convert OrientedBBox from subpixel precision to pixel precision by subtracting a subpixel offset from the coordinates.

        Points order in json format: [[left, top], [right, bottom]]

        In the labeling tool, labels are created with subpixel precision,
        which means that the coordinates of the oriented bounding box corners (top, left and bottom, right) can have decimal values representing fractions of a pixel.
        However, in Supervisely SDK, geometry coordinates are represented using pixel precision, where the coordinates are integers representing whole pixels.

        Example:
        Step 1. Input coordinates:
        - top = 1.55, left = 1.74, bottom = 4.63, right = 3.76

        Step 2. Round the coordinates (still remain in subpixel precision):
        - top = 1, left = 2, bottom = 5, right = 4
        - top will be rounded down to 2, left will be rounded down to 2, bottom will be rounded up to 6, right will be rounded down to 6

        Draw coordinates in pixel coordinate system:
            0   1   2   3   4   5
        0   +---+---+---+---+---+
            |   |   |   |   |   |
        1   +---+---+---+---+---+
            |   |   | x | x |   |
        2   +---+---+---+---+---+
            |   |   | x | x |   |
        3   +---+---+---+---+---+
            |   |   | x | x |   |
        4   +---+---+---+---+---+
            |   |   | x | x |   |
        5   +---+---+---+---+---+
                      x   x

        Step 3. Convert to pixel coordinates by subtracting a subpixel offset:
        - top = 1, left = 2, bottom = 4, right = 3

        Draw coordinates in pixel coordinate system:
            0   1   2   3   4   5
        0   +---+---+---+---+---+
            |   |   |   |   |   |
        1   +---+---+---+---+---+
            |   |   | x | x |   |
        2   +---+---+---+---+---+
            |   |   | x | x |   |
        3   +---+---+---+---+---+
            |   |   | x | x |   |
        4   +---+---+---+---+---+
            |   |   | x | x |   |
        5   +---+---+---+---+---+

        :param data: Json data with geometry config.
        :type data: :class:`dict`
        :param image_size: Image size in pixels (height, width).
        :type image_size: List[int]
        :return: Json data with coordinates converted to pixel coordinate system.
        :rtype: :class:`dict`
        """
        data = deepcopy(data)  # Avoid modifying the original data

        points = data[POINTS]
        [top, bottom] = sorted([points[0][1], points[1][1]])
        [left, right] = sorted([points[0][0], points[1][0]])

        top, left, bottom, right = cls._round_subpixel_coordinates(top, left, bottom, right)
        right = max(left, right - 1)
        bottom = max(top, bottom - 1)
        data[POINTS] = [[left, top], [right, bottom]]
        return data

    @classmethod
    def _to_subpixel_coordinate_system_json(cls, data: Dict) -> Dict:
        """
        Convert OrientedBBox from pixel precision to subpixel precision by adding a subpixel offset to the coordinates.

        Points order in json format: [[left, top], [right, bottom]]

        In the labeling tool, labels are created with subpixel precision,
        which means that the coordinates of the oriented bounding box corners (top, left and bottom, right) can have decimal values representing fractions of a pixel.
        However, in Supervisely SDK, geometry coordinates are represented using pixel precision, where the coordinates are integers representing whole pixels.

        :param data: Json data with geometry config.
        :type data: :class:`dict`
        :return: Json data with coordinates converted to subpixel coordinate system.
        :rtype: :class:`dict`
        """
        data = deepcopy(data)  # Avoid modifying the original data

        points = data[POINTS]
        [top, bottom] = sorted([points[0][1], points[1][1]])
        [left, right] = sorted([points[0][0], points[1][0]])

        right = max(left, right + 1)
        bottom = max(top, bottom + 1)
        data[POINTS] = [[left, top], [right, bottom]]
        return data



import json
import time
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional

import yaml

from supervisely._utils import is_development, logger
from supervisely.api.api import Api
from supervisely.api.image_api import ImageInfo
from supervisely.api.video.video_api import VideoInfo
from supervisely.app.widgets import Button, Card, Container, Stepper, Widget
from supervisely.geometry.any_geometry import AnyGeometry
from supervisely.io import env
from supervisely.nn.inference.inference import update_meta_and_ann_for_video_annotation
from supervisely.nn.inference.predict_app.gui.classes_selector import ClassesSelector
from supervisely.nn.inference.predict_app.gui.input_selector import InputSelector
from supervisely.nn.inference.predict_app.gui.model_selector import ModelSelector
from supervisely.nn.inference.predict_app.gui.output_selector import OutputSelector
from supervisely.nn.inference.predict_app.gui.settings_selector import (
    AddPredictionsMode,
    SettingsSelector,
)
from supervisely.nn.inference.predict_app.gui.tags_selector import TagsSelector
from supervisely.nn.inference.predict_app.gui.utils import (
    copy_items_to_project,
    create_project,
    disable_enable,
    update_custom_button_params,
    video_annotation_from_predictions,
)
from supervisely.nn.model.model_api import ModelAPI
from supervisely.nn.model.prediction import Prediction
from supervisely.project.project_meta import ProjectMeta
from supervisely.project.project_type import ProjectType
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_annotation import VideoAnnotation


class StepFlow:
    def __init__(self):
        self._stepper = None
        self.steps = {}
        self.steps_sequence = []

    def add_step(
        self,
        name: str,
        widget: Widget,
        on_select: Optional[Callable] = None,
        on_reactivate: Optional[Callable] = None,
        depends_on: Optional[List[Widget]] = None,
        on_lock: Optional[Callable] = None,
        on_unlock: Optional[Callable] = None,
        button: Optional[Button] = None,
        position: Optional[int] = None,
    ):
        if depends_on is None:
            depends_on = []
        self.steps[name] = {
            "widget": widget,
            "on_select": on_select,
            "on_reactivate": on_reactivate,
            "depends_on": depends_on,
            "on_lock": on_lock,
            "on_unlock": on_unlock,
            "button": button,
            "is_selected": False,
            "is_locked": False,
        }
        if button is not None:
            self._wrap_button(button, name)
        if position is not None:
            self.steps_sequence.insert(position, name)
        else:
            self.steps_sequence.append(name)
        self.update_locks()

    def _create_stepper(self):
        widgets = []
        for step_name in self.steps_sequence:
            step = self.steps[step_name]
            widgets.append(step["widget"])
        self._stepper = Stepper(widgets=widgets)

    @property
    def stepper(self):
        if self._stepper is None:
            self._create_stepper()
        return self._stepper

    def update_stepper(self):
        for i, step_name in enumerate(self.steps_sequence):
            step = self.steps[step_name]
            if not step["is_selected"]:
                self.stepper.set_active_step(i + 1)
                return

    def update_locks(self):
        for step in self.steps.values():
            should_lock = False
            for dep_name in step["depends_on"]:
                dep = self.steps[dep_name]
                if not dep["is_selected"]:
                    should_lock = True
                    break
            if should_lock and not step["is_locked"]:
                if step["on_lock"] is not None:
                    step["on_lock"]()
                step["is_locked"] = True
            if not should_lock and step["is_locked"]:
                if step["on_unlock"]:
                    step["on_unlock"]()
                step["is_locked"] = False

    def _reactivate_dependents(self, step_name: str, visited=None):
        if visited is None:
            visited = set()
        for dep_name, step in self.steps.items():
            if step_name in step["depends_on"] and not dep_name in visited:
                self._reactivate_step(dep_name, visited)

    def _reactivate_step(self, step_name: str, visited=None):
        step = self.steps[step_name]
        if step["on_reactivate"] is not None:
            step["on_reactivate"]()
        step["is_selected"] = False
        if visited is None:
            visited = set()
        self._reactivate_dependents(step_name, visited)

    def reactivate_step(self, step_name: str):
        self._reactivate_step(step_name)
        self.update_stepper()
        self.update_locks()

    def select_step(self, step_name: str):
        step = self.steps[step_name]
        if step["on_select"] is not None:
            step["on_select"]()
        step["is_selected"] = True
        self.update_stepper()
        self.update_locks()

    def select_or_reactivate(self, step_name: str):
        step = self.steps[step_name]
        if step["is_selected"]:
            self.reactivate_step(step_name)
        else:
            self.select_step(step_name)

    def _wrap_button(self, button: Button, step_name: str):
        button.click(lambda: self.select_or_reactivate(step_name))


class PredictAppGui:
    def __init__(self, api: Api, static_dir: str = "static"):
        self.api = api
        self.static_dir = static_dir

        # Environment variables
        self.team_id = env.team_id()
        self.workspace_id = env.workspace_id()
        self.project_id = env.project_id(raise_not_found=False)
        self.project_meta = None
        if self.project_id:
            self.project_meta = ProjectMeta.from_json(self.api.project.get_meta(self.project_id))
        # -------------------------------- #

        # Flags
        self._stop_flag = False
        self._is_running = False
        # -------------------------------- #

        # GUI
        # Steps
        self.step_flow = StepFlow()
        select_params = {"icon": None, "plain": False, "text": "Select"}
        reselect_params = {"icon": "zmdi zmdi-refresh", "plain": True, "text": "Reselect"}

        # 1. Input selector
        self.input_selector = InputSelector(self.workspace_id, self.api)

        def _on_input_select():
            valid = self.input_selector.validate_step()
            if not valid:
                return
            current_item_type = self.input_selector.radio.get_value()
            self.update_item_type()
            if self.model_api:
                if current_item_type == self.input_selector.radio.get_value():
                    inference_settings = self.model_api.get_settings()
                    self.settings_selector.set_inference_settings(inference_settings)

                    if self.input_selector.radio.get_value() == ProjectType.VIDEOS.value:
                        try:
                            tracking_settings = self.model_api.get_tracking_settings()
                            self.settings_selector.set_tracking_settings(tracking_settings)
                        except Exception as e:
                            logger.warning(
                                "Unable to get tracking settings from the model. Settings defaults"
                            )
                            self.settings_selector.set_default_tracking_settings()
            self.input_selector.disable()

            self.project_id = self.input_selector.get_project_id()
            if self.project_id:
                self.project_meta = ProjectMeta.from_json(self.api.project.get_meta(self.project_id))
            update_custom_button_params(self.input_selector.button, reselect_params)

        def _on_input_reactivate():
            self.input_selector.enable()
            update_custom_button_params(self.input_selector.button, select_params)

        self.step_flow.add_step(
            name="input_selector",
            widget=self.input_selector.card,
            on_select=_on_input_select,
            on_reactivate=_on_input_reactivate,
            button=self.input_selector.button,
        )

        # 2. Model selector
        self.model_selector = ModelSelector(self.api, self.team_id)

        self.step_flow.add_step(
            name="model_selector",
            widget=self.model_selector.card,
        )

        # 3. Classes selector
        self.classes_selector = ClassesSelector()

        def _on_classes_select():
            valid = self.classes_selector.validate_step()
            if not valid:
                return
            self.classes_selector.classes_table.disable()

            # Find conflict between project meta and model meta
            selected_classes_names = self.classes_selector.get_selected_classes()
            project_meta = self.project_meta
            model_meta = self.model_api.get_model_meta()

            has_conflict = False
            for class_name in selected_classes_names:
                project_obj_class = project_meta.get_obj_class(class_name)
                if project_obj_class is None:
                    continue

                model_obj_class = model_meta.get_obj_class(class_name)
                if model_obj_class.geometry_type.name() == AnyGeometry.name():
                    continue

                if project_obj_class.geometry_type.name() == model_obj_class.geometry_type.name():
                    continue

                has_conflict = True
                break

            if has_conflict:
                self.settings_selector.model_prediction_suffix_container.show()
            else:
                self.settings_selector.model_prediction_suffix_container.hide()
            # ------------------------------------------------ #

            update_custom_button_params(self.classes_selector.button, reselect_params)

        def _on_classes_reactivate():
            self.classes_selector.classes_table.enable()
            update_custom_button_params(self.classes_selector.button, select_params)

        self.step_flow.add_step(
            name="classes_selector",
            widget=self.classes_selector.card,
            on_select=_on_classes_select,
            on_reactivate=_on_classes_reactivate,
            depends_on=["input_selector", "model_selector"],
            on_lock=self.classes_selector.lock,
            on_unlock=self.classes_selector.unlock,
            button=self.classes_selector.button,
        )

        # 4. Tags selector
        self.tags_selector = None
        if False:
            self.tags_selector = TagsSelector()
            self.step_flow.add_step("tags_selector", self.tags_selector.card)

        # 5. Settings selector & Preview
        self.settings_selector = SettingsSelector(
            api=self.api,
            static_dir=self.static_dir,
            model_selector=self.model_selector,
            input_selector=self.input_selector,
        )

        def _on_settings_select():
            valid = self.settings_selector.validate_step()
            if not valid:
                return
            self.settings_selector.disable()
            update_custom_button_params(self.settings_selector.button, reselect_params)

        def _on_settings_reactivate():
            self.settings_selector.enable()
            update_custom_button_params(self.settings_selector.button, select_params)

        self.step_flow.add_step(
            name="settings_selector",
            widget=self.settings_selector.cards_container,
            on_select=_on_settings_select,
            on_reactivate=_on_settings_reactivate,
            depends_on=["input_selector", "model_selector", "classes_selector"],
            on_lock=self.settings_selector.lock,
            on_unlock=self.settings_selector.unlock,
            button=self.settings_selector.button,
        )
        self.settings_selector.preview.run_button.disable()

        # 6. Output selector
        self.output_selector = OutputSelector(self.api)

        self.step_flow.add_step(
            "output_selector",
            self.output_selector.card,
            depends_on=[
                "input_selector",
                "model_selector",
                "classes_selector",
                # "tags_selector",
                "settings_selector",
            ],
            on_lock=self.output_selector.lock,
            on_unlock=self.output_selector.unlock,
        )
        # -------------------------------- #

        # Layout
        self.layout = Container([self.step_flow.stepper])
        # ---------------------------- #

        def set_entity_meta():
            model_api = self.model_selector.model.model_api

            model_meta = model_api.get_model_meta()
            if self.classes_selector is not None:
                self.classes_selector.set_project_meta(model_meta)
                self.classes_selector.classes_table.show()
            if self.tags_selector is not None:
                self.tags_selector.tags_table.set_project_meta(model_meta)
                self.tags_selector.tags_table.show()

            inference_settings = model_api.get_settings()
            self.settings_selector.set_inference_settings(inference_settings)

            if self.input_selector.radio.get_value() == ProjectType.VIDEOS.value:
                try:
                    tracking_settings = model_api.get_tracking_settings()
                    self.settings_selector.set_tracking_settings(tracking_settings)
                except Exception as e:
                    logger.warning(
                        "Unable to get tracking settings from the model. Settings defaults"
                    )
                    self.settings_selector.set_default_tracking_settings()

        def reset_entity_meta():
            empty_meta = ProjectMeta()
            if self.classes_selector is not None:
                self.classes_selector.set_project_meta(empty_meta)
                self.classes_selector.classes_table.hide()
            if self.tags_selector is not None:
                self.tags_selector.tags_table.set_project_meta(empty_meta)
                self.tags_selector.tags_table.hide()

            self.settings_selector.set_inference_settings("")

        def deploy_and_set_step():
            self.model_selector.validator_text.hide()
            model_api = type(self.model_selector.model).deploy(self.model_selector.model)
            if model_api is not None:
                set_entity_meta()
                self.step_flow.select_step("model_selector")
            else:
                reset_entity_meta()
                self.step_flow.reactivate_step("model_selector")
            return model_api

        def stop_and_reset_step():
            type(self.model_selector.model).stop(self.model_selector.model)
            self.step_flow.reactivate_step("model_selector")
            reset_entity_meta()

        def disconnect_and_reset_step():
            type(self.model_selector.model).disconnect(self.model_selector.model)
            self.step_flow.reactivate_step("model_selector")
            reset_entity_meta()

        # Replace deploy methods for DeployModel widget
        self.model_selector.model.deploy = deploy_and_set_step
        self.model_selector.model.stop = stop_and_reset_step
        self.model_selector.model.disconnect = disconnect_and_reset_step

        # ------------------------------------------------- #

    @property
    def model_api(self) -> Optional[ModelAPI]:
        return self.model_selector.model.model_api

    def update_item_type(self):
        item_type = self.input_selector.radio.get_value()
        self.settings_selector.update_item_type(item_type)
        self.output_selector.update_item_type(item_type)

    def _run_videos(self, run_parameters: Dict[str, Any]) -> List[Prediction]:
        if self.model_api is None:
            self.set_validator_text("Deploying model...", "info")
            self.model_selector.model._deploy()
        if self.model_api is None:
            logger.error("Model Deployed with an error")
            raise RuntimeError("Model Deployed with an error")

        self.set_validator_text("Preparing settings for prediction...", "info")
        if run_parameters is None:
            run_parameters = self.get_run_parameters()

        input_parameters = run_parameters["input"]
        input_video_ids = input_parameters["video_ids"]
        if not input_video_ids:
            raise ValueError("No video IDs provided for video prediction.")

        predict_kwargs = {}
        # Settings
        settings = run_parameters["settings"]
        model_prediction_suffix = settings.pop("model_prediction_suffix", "")
        prediction_mode = settings.pop("predictions_mode")
        tracking = settings.pop("tracking", False)
        predict_kwargs.update(settings)

        # Classes
        classes = run_parameters["classes"]
        if classes:
            predict_kwargs["classes"] = classes

        output_parameters = run_parameters["output"]
        project_name = output_parameters.get("project_name", "")
        upload_to_source_project = output_parameters.get("upload_to_source_project", False)
        skip_project_versioning = output_parameters.get("skip_project_versioning", False)
        skip_annotated = output_parameters.get("skip_annotated", False)

        video_infos_by_project_id: Dict[int, List[VideoInfo]] = {}
        video_infos_by_dataset_id: Dict[int, List[VideoInfo]] = {}
        for info in self.api.video.get_info_by_id_batch(input_video_ids):
            video_infos_by_project_id.setdefault(info.project_id, []).append(info)
            video_infos_by_dataset_id.setdefault(info.dataset_id, []).append(info)
        src_project_metas: Dict[int, ProjectMeta] = {}
        for project_id in video_infos_by_project_id.keys():
            src_project_metas[project_id] = ProjectMeta.from_json(
                self.api.project.get_meta(project_id)
            )

        video_ids_to_skip = set()
        if skip_annotated:
            self.set_validator_text("Checking for already annotated videos...", "info")
            secondary_pbar = self.output_selector.secondary_progress(
                message="Checking for already annotated videos...", total=len(input_video_ids)
            )
            self.output_selector.secondary_progress.show()
            for dataset_id, video_infos in video_infos_by_dataset_id.items():
                annotations = self.api.video.annotation.download_bulk(
                    dataset_id, [info.id for info in video_infos]
                )
                for ann_json, video_info in zip(annotations, video_infos):
                    if ann_json:
                        project_meta = src_project_metas[video_info.project_id]
                        ann = VideoAnnotation.from_json(ann_json, project_meta=project_meta)
                        if len(ann.figures) > 0:
                            video_ids_to_skip.add(video_info.id)
                    secondary_pbar.update()
            self.output_selector.secondary_progress.hide()
            if video_ids_to_skip:
                video_infos_by_project_id = {
                    pid: [info for info in infos if info.id not in video_ids_to_skip]
                    for pid, infos in video_infos_by_project_id.items()
                }

        main_pbar_str = "Processing videos..."
        if video_ids_to_skip:
            main_pbar_str += f" (Skipped {len(video_ids_to_skip)} already annotated videos)"
        total_videos = sum(len(v) for v in video_infos_by_project_id.values())
        if total_videos == 0:
            self.set_validator_text(
                f"No videos to process. Skipped {len(video_ids_to_skip)} already annotated videos",
                "warning",
            )
            return []
        main_pbar = self.output_selector.progress(message=main_pbar_str, total=total_videos)
        self.output_selector.progress.show()
        all_predictictions: List[Prediction] = []
        for src_project_id, src_video_infos in video_infos_by_project_id.items():
            if len(src_video_infos) == 0:
                continue
            project_info = self.api.project.get_info_by_id(src_project_id)
            project_validator_text_str = (
                f"Processing project: {project_info.name} [id: {src_project_id}]"
            )
            if upload_to_source_project:
                if not skip_project_versioning and not is_development():
                    logger.info("Creating new project version...")
                    self.set_validator_text(
                        project_validator_text_str + ": Creating project version",
                        "info",
                    )
                    version_id = self.api.project.version.create(
                        project_info,
                        "Created by Predict App. Task Id: " + str(env.task_id()),
                    )
                    logger.info("New project version created: " + str(version_id))
                output_project_id = src_project_id
                output_videos: List[VideoInfo] = src_video_infos
            else:
                self.set_validator_text(
                    project_validator_text_str + ": Creating project...", "info"
                )
                if not project_name:
                    project_name = project_info.name + " [Predictions]"
                    logger.warning(
                        "Project name is empty, using auto-generated name: " + project_name
                    )
                with_annotations = prediction_mode in [
                    AddPredictionsMode.APPEND,
                    AddPredictionsMode.IOU_MERGE,
                ]
                created_project = create_project(
                    api=self.api,
                    project_id=src_project_id,
                    project_name=project_name,
                    workspace_id=self.workspace_id,
                    copy_meta=with_annotations,
                    project_type=ProjectType.VIDEOS,
                )
                output_project_id = created_project.id
                output_videos: List[VideoInfo] = copy_items_to_project(
                    api=self.api,
                    src_project_id=src_project_id,
                    items=src_video_infos,
                    dst_project_id=created_project.id,
                    with_annotations=with_annotations,
                    ds_progress=self.output_selector.secondary_progress,
                    project_type=ProjectType.VIDEOS,
                )

            self.set_validator_text(
                project_validator_text_str + ": Merging project meta",
                "info",
            )
            project_meta = src_project_metas[src_project_id]
            for src_video_info, output_video_info in zip(src_video_infos, output_videos):
                video_validator_text_str = (
                    project_validator_text_str
                    + f", video: {src_video_info.name} [id: {src_video_info.id}]"
                )
                self.set_validator_text(
                    video_validator_text_str + ": Predicting",
                    "info",
                )
                frames_predictions: List[Prediction] = []
                with self.model_api.predict_detached(
                    video_id=src_video_info.id,
                    tqdm=self.output_selector.secondary_progress(),
                    tracking=tracking,
                    **predict_kwargs,
                ) as session:
                    self.output_selector.secondary_progress.show()
                    for prediction in session:
                        if self._stop_flag:
                            logger.info("Prediction stopped by user.")
                            raise StopIteration("Stopped by user.")
                        frames_predictions.append(prediction)
                    all_predictictions.extend(frames_predictions)
                    if tracking:
                        prediction_video_annotation: VideoAnnotation = VideoAnnotation.from_json(
                            session.final_result["video_ann"],
                            project_meta=self.model_api.get_model_meta(),
                        )
                    else:
                        prediction_video_annotation = video_annotation_from_predictions(
                            frames_predictions,
                            project_meta,
                            frame_size=(src_video_info.frame_height, src_video_info.frame_width),
                        )
                if prediction_video_annotation is None:
                    logger.warning(
                        f"No predictions were made for video {src_video_info.name} [id: {src_video_info.id}]"
                    )
                    main_pbar.update()
                    continue
                self.set_validator_text(
                    video_validator_text_str + ": Uploading predictions",
                    "info",
                )
                project_meta, prediction_video_annotation, meta_changed = (
                    update_meta_and_ann_for_video_annotation(
                        meta=project_meta,
                        ann=prediction_video_annotation,
                        model_prediction_suffix=model_prediction_suffix,
                    )
                )
                if meta_changed:
                    self.api.project.update_meta(output_project_id, project_meta)
                if upload_to_source_project:
                    if prediction_mode in [
                        AddPredictionsMode.REPLACE,
                        AddPredictionsMode.REPLACE_EXISTING_LABELS_AND_SAVE_IMAGE_TAGS,
                    ]:
                        self.output_selector.secondary_progress.hide()
                        with open("/tmp/prediction_video_annotation.json", "w") as f:
                            json.dump(prediction_video_annotation.to_json(), f)
                        self.api.video.annotation.upload_paths(
                            video_ids=[src_video_info.id],
                            paths=["/tmp/prediction_video_annotation.json"],
                            project_meta=project_meta,
                        )
                    else:
                        secondary_pbar = self.output_selector.secondary_progress(
                            message="Uploading annotations...",
                            total=len(prediction_video_annotation.figures),
                        )
                        self.output_selector.secondary_progress.show()
                        self.api.video.annotation.append(
                            video_id=src_video_info.id,
                            ann=prediction_video_annotation,
                            key_id_map=KeyIdMap(),
                            progress_cb=secondary_pbar.update,
                        )
                else:
                    secondary_pbar = self.output_selector.secondary_progress(
                        message="Uploading annotations...",
                        total=len(prediction_video_annotation.figures),
                    )
                    self.output_selector.secondary_progress.show()
                    self.api.video.annotation.append(
                        video_id=output_video_info.id,
                        ann=prediction_video_annotation,
                        key_id_map=KeyIdMap(),
                        progress_cb=secondary_pbar.update,
                    )
                main_pbar.update()
        self.set_validator_text("Project successfully processed", "success")
        self.output_selector.set_result_thumbnail(output_project_id)
        return all_predictictions

    def _run_images(self, run_parameters: Dict[str, Any] = None) -> List[Prediction]:
        if self.model_api is None:
            self.set_validator_text("Deploying model...", "info")
            self.model_selector.model._deploy()
            if isinstance(run_parameters, dict):
                # Set classes from run_parameters
                model_classes = self.model_api.get_classes()
                classes = run_parameters.get("classes", [])
                classes = set(classes) & set(model_classes)
                self.classes_selector.set_classes(list(classes))

                # Set inference settings from run_parameters
                settings = run_parameters.get("settings", {})
                inference_settings = settings.get("inference_settings")
                if inference_settings:
                    self.settings_selector.set_inference_settings(inference_settings)

        if self.model_api is None:
            logger.error("Model Deployed with an error")
            raise RuntimeError("Model Deployed with an error")

        self.set_validator_text("Preparing settings for prediction...", "info")
        if run_parameters is None:
            run_parameters = self.get_run_parameters()

        predict_kwargs = {}
        # Input
        input_args = {}
        input_parameters = run_parameters["input"]
        input_project_id = input_parameters.get("project_id", None)
        input_dataset_ids = input_parameters.get("dataset_ids", [])
        input_image_ids = input_parameters.get("image_ids", [])
        if input_image_ids:
            input_args["image_ids"] = input_image_ids
        elif input_dataset_ids:
            input_args["dataset_ids"] = input_dataset_ids
        elif input_project_id:
            input_args["project_id"] = input_project_id
        else:
            raise ValueError("No valid input parameters found for prediction.")

        # Settings
        settings = run_parameters["settings"]
        prediction_mode = settings.pop("predictions_mode")
        upload_mode = None
        with_annotations = None
        if prediction_mode == AddPredictionsMode.REPLACE:
            upload_mode = "replace"
            with_annotations = False
        elif prediction_mode == AddPredictionsMode.APPEND:
            upload_mode = "append"
            with_annotations = True
        elif prediction_mode == AddPredictionsMode.IOU_MERGE:
            upload_mode = "iou_merge"
            with_annotations = True
        elif prediction_mode == AddPredictionsMode.REPLACE_EXISTING_LABELS_AND_SAVE_IMAGE_TAGS:
            upload_mode = "replace"
            with_annotations = True
        predict_kwargs.update(settings)
        predict_kwargs["upload_mode"] = upload_mode

        # Classes
        classes = run_parameters["classes"]
        if classes:
            predict_kwargs["classes"] = classes

        # Output
        output_parameters = run_parameters["output"]
        project_name = output_parameters.get("project_name", None)
        upload_to_source_project = output_parameters.get("upload_to_source_project", False)
        skip_project_versioning = output_parameters.get("skip_project_versioning", False)
        skip_annotated = output_parameters.get("skip_annotated", False)

        image_infos = []
        if input_image_ids:
            image_infos = self.api.image.get_info_by_id_batch(input_image_ids)
        elif input_dataset_ids:
            for dataset_id in input_dataset_ids:
                image_infos.extend(self.api.image.get_list(dataset_id))
        elif input_project_id:
            datasets = self.api.dataset.get_list(input_project_id, recursive=True)
            for dataset in datasets:
                image_infos.extend(self.api.image.get_list(dataset.id))
        if len(image_infos) == 0:
            raise ValueError("No images found for the given input parameters.")

        to_skip = []
        if skip_annotated:
            to_skip = [image_info.id for image_info in image_infos if image_info.labels_count == 0]
        if to_skip:
            image_infos = [info for info in image_infos if info.id not in to_skip]
        if len(image_infos) == 0:
            self.set_validator_text(
                f"All images are already annotated. Nothing to predict.", "warning"
            )
            return []

        image_infos_by_project_id: Dict[int, List[ImageInfo]] = {}
        image_infos_by_dataset_id: Dict[int, List[ImageInfo]] = {}
        ds_project_mapping: Dict[int, int] = {}
        for info in image_infos:
            image_infos_by_dataset_id.setdefault(info.dataset_id, []).append(info)
            if info.dataset_id not in ds_project_mapping:
                ds_info = self.api.dataset.get_info_by_id(info.dataset_id)
                ds_project_mapping[info.dataset_id] = ds_info.project_id
            project_id = ds_project_mapping[info.dataset_id]
            image_infos_by_project_id.setdefault(project_id, []).append(info)

        src_project_metas: Dict[int, ProjectMeta] = {}
        for project_id in image_infos_by_project_id.keys():
            src_project_metas[project_id] = ProjectMeta.from_json(
                self.api.project.get_meta(project_id)
            )

        self.output_selector.progress.show()
        total_items = sum(len(v) for v in image_infos_by_project_id.values())
        main_pbar = self.output_selector.progress(message=f"Copying images...", total=total_items)
        for src_project_id, infos in image_infos_by_project_id.items():
            if len(infos) == 0:
                continue
            project_info = self.api.project.get_info_by_id(src_project_id)
            project_validator_text_str = (
                f"Processing project: {project_info.name} [id: {src_project_id}]"
            )
            if upload_to_source_project:
                if not skip_project_versioning and not is_development():
                    logger.info("Creating new project version...")
                    self.set_validator_text(
                        project_validator_text_str + ": Creating project version", "info"
                    )
                    version_id = self.api.project.version.create(
                        project_info,
                        "Created by Predict App. Task Id: " + str(env.task_id()),
                    )
                    logger.info("New project version created: " + str(version_id))
                output_project_id = src_project_id
                output_image_infos: List[ImageInfo] = infos
            else:
                self.set_validator_text(
                    project_validator_text_str + ": Creating project...", "info"
                )
                if not project_name:
                    project_name = project_info.name + " [Predictions]"
                    logger.warning(
                        "Project name is empty, using auto-generated name: " + project_name
                    )
                created_project = create_project(
                    api=self.api,
                    project_id=src_project_id,
                    project_name=project_name,
                    workspace_id=self.workspace_id,
                    copy_meta=with_annotations,
                    project_type=ProjectType.IMAGES,
                )
                output_project_id = created_project.id
                output_image_infos: List[ImageInfo] = copy_items_to_project(
                    api=self.api,
                    src_project_id=src_project_id,
                    items=infos,
                    dst_project_id=created_project.id,
                    with_annotations=with_annotations,
                    ds_progress=self.output_selector.secondary_progress,
                    progress_cb=main_pbar.update,
                    project_type=ProjectType.IMAGES,
                )

        # Run prediction
        self.set_validator_text("Running prediction...", "info")
        predictions: List[Prediction] = []
        self._is_running = True
        with self.model_api.predict_detached(
            image_ids=[info.id for info in output_image_infos],
            **predict_kwargs,
            tqdm=self.output_selector.progress(),
        ) as session:
            for prediction in session:
                if self._stop_flag:
                    logger.info("Prediction stopped by user.")
                    raise StopIteration("Stopped by user.")
                predictions.append(prediction)
        self.set_validator_text("Project successfully processed", "success")
        self.output_selector.set_result_thumbnail(output_project_id)
        return predictions

    def run(self, run_parameters: Dict[str, Any] = None) -> List[Prediction]:
        self.show_validator_text()
        if run_parameters is None:
            run_parameters = self.get_run_parameters()
        input_parameters = run_parameters["input"]
        video_ids = input_parameters.get("video_ids", None)
        try:
            if video_ids:
                run_f = self._run_videos
            else:
                run_f = self._run_images
            return run_f(run_parameters)
        except StopIteration:
            logger.info("Prediction stopped by user.")
            self.set_validator_text("Prediction stopped by user.", "warning")
            raise
        except Exception as e:
            logger.error(f"Error during prediction: {str(e)}")
            self.set_validator_text(f"Error during prediction: {str(e)}", "error")
            disable_enable(self.output_selector.widgets_to_disable, False)
            raise
        finally:
            self.output_selector.secondary_progress.hide()
            self.output_selector.progress.hide()
            self._is_running = False
            self._stop_flag = False

    def stop(self):
        logger.info("Stopping prediction...")
        self._stop_flag = True

    def wait_for_stop(self, timeout: int = None):
        logger.info(
            "Waiting " + ""
            if timeout is None
            else f"{timeout} seconds " + "for prediction to stop..."
        )
        t = time.monotonic()
        while self._is_running:
            if timeout is not None and time.monotonic() - t > timeout:
                raise TimeoutError("Timeout while waiting for stop.")
            time.sleep(0.1)
        logger.info("Prediction stopped.")

    def shutdown_model(self):
        self.stop()
        self.wait_for_stop(10)
        self.model_selector.model.stop()

    def get_run_parameters(self) -> Dict[str, Any]:
        settings = {
            "model": self.model_selector.model.get_deploy_parameters(),
            "settings": self.settings_selector.get_settings(),
            "input": self.input_selector.get_settings(),
            "output": self.output_selector.get_settings(),
        }
        if self.classes_selector is not None:
            settings["classes"] = self.classes_selector.get_selected_classes()
        if self.tags_selector is not None:
            settings["tags"] = self.tags_selector.get_selected_tags()
        return settings

    def load_from_json(self, data):
        # 1. Input selector
        self.input_selector.load_from_json(data.get("input", {}))
        # self.input_selector_cb()

        # 2. Model selector
        self.model_selector.model.load_from_json(data.get("model", {}))
        # TODO: deploy model here?

        # 3. Classes selector
        if self.classes_selector is not None:
            self.classes_selector.load_from_json(data.get("classes", {}))

        # 4. Tags selector
        if self.tags_selector is not None:
            self.tags_selector.load_from_json(data.get("tags", {}))

        # 5. Settings selector & Preview
        self.settings_selector.load_from_json(data.get("settings", {}))

        # 6. Output selector
        self.output_selector.load_from_json(data.get("output", {}))

    def set_validator_text(self, text: str, status: str = "text"):
        self.output_selector.validator_text.set(text=text, status=status)

    def show_validator_text(self):
        self.output_selector.validator_text.show()

    def hide_validator_text(self):
        self.output_selector.validator_text.hide()



import os
from collections import defaultdict
from typing import Dict, List, Literal, Optional, Tuple, Union

import supervisely as sly
from supervisely._utils import rand_str
from supervisely.api.api import Api
from supervisely.app import DataJson, StateJson, get_data_dir
from supervisely.app.widgets import (
    Container,
    Field,
    NotificationBox,
    RadioTabs,
    SelectString,
    SelectTagMeta,
    Widget,
)
from supervisely.app.widgets.random_splits_table.random_splits_table import (
    RandomSplitsTable,
)
from supervisely.app.widgets.select_collection.select_collection import SelectCollection
from supervisely.app.widgets.select_dataset_tree.select_dataset_tree import (
    SelectDatasetTree,
)
from supervisely.io.fs import remove_dir
from supervisely.project import get_project_class
from supervisely.project.pointcloud_episode_project import PointcloudEpisodeProject
from supervisely.project.pointcloud_project import PointcloudProject
from supervisely.project.project import ItemInfo, Project
from supervisely.project.project_type import ProjectType
from supervisely.project.video_project import VideoProject
from supervisely.project.volume_project import VolumeProject


class TrainValSplits(Widget):
    def __init__(
        self,
        project_id: Optional[int] = None,
        project_fs: Optional[
            Union[Project, VideoProject, VolumeProject, PointcloudProject, PointcloudEpisodeProject]
        ] = None,
        random_splits: Optional[bool] = True,
        tags_splits: Optional[bool] = True,
        datasets_splits: Optional[bool] = True,
        widget_id: Optional[int] = None,
        collections_splits: Optional[bool] = False,
    ):
        self._project_id = project_id
        self._project_fs = project_fs

        self._project_info = None
        self._project_type = None
        self._project_class = None
        self._api = None
        if project_id is not None:
            self._api = Api()
            self._project_info = self._api.project.get_info_by_id(
                self._project_id, raise_error=True
            )

        if project_fs is not None:
            self._project_type = project_fs.type
        elif self._project_info is not None:
            self._project_type = self._project_info.type
        if self._project_type is not None:
            self._project_class = get_project_class(self._project_type)

        self._random_splits_table: RandomSplitsTable = None
        self._train_tag_select: SelectTagMeta = None
        self._val_tag_select: SelectTagMeta = None
        self._untagged_select: SelectString = None
        self._train_ds_select: Union[SelectDatasetTree, SelectString] = None
        self._val_ds_select: Union[SelectDatasetTree, SelectString] = None
        self._train_collections_select: SelectCollection = None
        self._val_collections_select: SelectCollection = None
        self._split_methods = []

        contents = []
        tabs_descriptions = []
        if random_splits:
            self._split_methods.append("Random")
            tabs_descriptions.append("Shuffle data and split with defined probability")
            contents.append(self._get_random_content())
        if tags_splits:
            self._split_methods.append("Based on item tags")
            tabs_descriptions.append(
                f"{self._project_type.capitalize()} should have assigned train or val tag"
            )
            contents.append(self._get_tags_content())
        if datasets_splits:
            self._split_methods.append("Based on datasets")
            tabs_descriptions.append("Select one or several datasets for every split")
            contents.append(self._get_datasets_content())
        if collections_splits:
            self._split_methods.append("Based on collections")
            tabs_descriptions.append("Select one or several collections for every split")
            contents.append(self._get_collections_content())
        if not self._split_methods:
            raise ValueError(
                "Any of split methods [random_splits, tags_splits, datasets_splits] must be specified in TrainValSplits."
            )

        self._content = RadioTabs(
            titles=self._split_methods,
            descriptions=tabs_descriptions,
            contents=contents,
        )

        super().__init__(widget_id=widget_id, file_path=__file__)

    def _get_random_content(self):
        items_count = 0
        if self._project_id is not None:
            items_count = self._project_info.items_count
        elif self._project_fs is not None:
            items_count = self._project_fs.total_items
        self._random_splits_table = RandomSplitsTable(items_count)

        return Container(widgets=[self._random_splits_table], direction="vertical", gap=5)

    def _get_tags_content(self):
        notification_box = NotificationBox(
            title="Notice: How to make equal splits",
            description="Choose the same tag for train/validation to make splits equal. Can be used for debug and for tiny projects",
            box_type="info",
        )
        if self._project_id is not None:
            self._train_tag_select = SelectTagMeta(project_id=self._project_id, show_label=False)
            self._val_tag_select = SelectTagMeta(project_id=self._project_id, show_label=False)
        elif self._project_fs is not None:
            self._train_tag_select = SelectTagMeta(
                project_meta=self._project_fs.meta, show_label=False
            )
            self._val_tag_select = SelectTagMeta(
                project_meta=self._project_fs.meta, show_label=False
            )
        self._untagged_select = SelectString(
            values=["train", "val", "ignore"],
            labels=[
                f"add untagged {self._project_type} to train set",
                f"add untagged {self._project_type} to val set",
                f"ignore untagged {self._project_type}",
            ],
            placeholder="Select action",
        )
        train_field = Field(
            self._train_tag_select,
            title="Train tag",
            description=f"all {self._project_type} with this tag are considered as training set",
        )
        val_field = Field(
            self._val_tag_select,
            title="Validation tag",
            description=f"all {self._project_type} with this tag are considered as validation set",
        )
        without_tags_field = Field(
            self._untagged_select,
            title=f"{self._project_type.capitalize()} without selected tags",
            description=f"Choose what to do with untagged {self._project_type}",
        )
        return Container(
            widgets=[
                notification_box,
                train_field,
                val_field,
                without_tags_field,
            ],
            direction="vertical",
            gap=5,
        )

    def _get_datasets_content(self):
        notification_box = NotificationBox(
            title="Notice: How to make equal splits",
            description="Choose the same dataset(s) for train/validation to make splits equal. Can be used for debug and for tiny projects",
            box_type="info",
        )
        if self._project_id is not None:
            self._train_ds_select = SelectDatasetTree(
                multiselect=True,
                flat=True,
                select_all_datasets=False,
                allowed_project_types=[self._project_type],
                always_open=False,
                compact=True,
                team_is_selectable=False,
                workspace_is_selectable=False,
                append_to_body=True,
            )

            self._val_ds_select = SelectDatasetTree(
                multiselect=True,
                flat=True,
                select_all_datasets=False,
                allowed_project_types=[self._project_type],
                always_open=False,
                compact=True,
                team_is_selectable=False,
                workspace_is_selectable=False,
                append_to_body=True,
            )

            # old implementation
            # self._train_ds_select = SelectDataset(
            #     project_id=self._project_id, multiselect=True, compact=True, show_label=False
            # )
            # self._val_ds_select = SelectDataset(
            #     project_id=self._project_id, multiselect=True, compact=True, show_label=False
            # )
        elif self._project_fs is not None:
            ds_names = [ds.name for ds in self._project_fs.datasets]
            self._train_ds_select = SelectString(ds_names, multiple=True)
            self._val_ds_select = SelectString(ds_names, multiple=True)
        train_field = Field(
            self._train_ds_select,
            title="Train dataset(s)",
            description=f"all {self._project_type} in selected dataset(s) are considered as training set",
        )
        val_field = Field(
            self._val_ds_select,
            title="Validation dataset(s)",
            description=f"all {self._project_type} in selected dataset(s) are considered as validation set",
        )
        return Container(
            widgets=[notification_box, train_field, val_field], direction="vertical", gap=5
        )

    def _get_collections_content(self):
        notification_box = NotificationBox(
            title="Notice: How to make equal splits",
            description="Choose the same collection(s) for train/validation to make splits equal. Can be used for debug and for tiny projects",
            box_type="info",
        )

        self._train_collections_select = SelectCollection(multiselect=True, compact=True)
        self._val_collections_select = SelectCollection(multiselect=True, compact=True)
        if self._project_id is not None:
            self._train_collections_select.set_project_id(self._project_id)
            self._val_collections_select.set_project_id(self._project_id)
        train_field = Field(
            self._train_collections_select,
            title="Train collection(s)",
            description="all images in selected collection(s) are considered as training set",
        )
        val_field = Field(
            self._val_collections_select,
            title="Validation collection(s)",
            description="all images in selected collection(s) are considered as validation set",
        )
        return Container(
            widgets=[notification_box, train_field, val_field], direction="vertical", gap=5
        )

    def get_json_data(self):
        return {}

    def get_json_state(self):
        return {}

    def get_splits(self) -> Tuple[List[ItemInfo], List[ItemInfo]]:
        if self._project_id is None and self._project_fs is None:
            raise ValueError("Both project_id and project_fs are None.")
        split_method = self._content.get_active_tab()
        tmp_project_dir = None
        train_set, val_set = [], []
        if self._project_fs is None:
            tmp_project_dir = os.path.join(get_data_dir(), rand_str(15))
            self._project_class.download(self._api, self._project_id, tmp_project_dir)

        project_dir = tmp_project_dir if tmp_project_dir is not None else self._project_fs.directory

        if split_method == "Random":
            splits_counts = self._random_splits_table.get_splits_counts()
            train_count = splits_counts["train"]
            val_count = splits_counts["val"]
            val_part = val_count / (val_count + train_count)
            project = self._project_class(project_dir, sly.OpenMode.READ)
            n_images = project.total_items
            new_val_count = round(val_part * n_images)
            new_train_count = n_images - new_val_count

            train_set, val_set = self._project_class.get_train_val_splits_by_count(
                project_dir, new_train_count, new_val_count
            )

        elif split_method == "Based on item tags":
            train_tag_name = self._train_tag_select.get_selected_name()
            val_tag_name = self._val_tag_select.get_selected_name()
            add_untagged_to = self._untagged_select.get_value()
            train_set, val_set = self._project_class.get_train_val_splits_by_tag(
                project_dir, train_tag_name, val_tag_name, add_untagged_to
            )

        elif split_method == "Based on datasets":
            if self._project_id is not None:
                self._train_ds_select: SelectDatasetTree
                self._val_ds_select: SelectDatasetTree
                train_ds_ids = self._train_ds_select.get_selected_ids()
                val_ds_ids = self._val_ds_select.get_selected_ids()
                ds_infos = [dataset for _, dataset in self._api.dataset.tree(self._project_id)]
                train_ds_names, val_ds_names = [], []
                for ds_info in ds_infos:
                    if ds_info.id in train_ds_ids:
                        train_ds_names.append(ds_info.name)
                    if ds_info.id in val_ds_ids:
                        val_ds_names.append(ds_info.name)
            elif self._project_fs is not None:
                self._train_ds_select: SelectString
                self._val_ds_select: SelectString
                train_ds_names = self._train_ds_select.get_value()
                val_ds_names = self._val_ds_select.get_value()

            train_set, val_set = self._project_class.get_train_val_splits_by_dataset(
                project_dir, train_ds_names, val_ds_names
            )
        elif split_method == "Based on collections":
            if self._project_id is None:
                raise ValueError(
                    "You can not use collections_splits parameter without project_id parameter."
                )
            train_collections = self._train_collections_select.get_selected_ids()
            val_collections = self._val_collections_select.get_selected_ids()

            train_set, val_set = self._project_class.get_train_val_splits_by_collections(
                project_dir,
                train_collections,
                val_collections,
                self._project_id,
                self._api,
            )

        if tmp_project_dir is not None:
            remove_dir(tmp_project_dir)
        return train_set, val_set

    def set_split_method(self, split_method: Literal["random", "tags", "datasets", "collections"]):
        if split_method == "random":
            split_method = "Random"
        elif split_method == "tags":
            split_method = "Based on item tags"
        elif split_method == "datasets":
            split_method = "Based on datasets"
        elif split_method == "collections":
            split_method = "Based on collections"
        self._content.set_active_tab(split_method)
        StateJson().send_changes()
        DataJson().send_changes()

    def get_split_method(self) -> str:
        return self._content.get_active_tab()

    def set_random_splits(
        self, split: Literal["train", "training", "val", "validation"], percent: int
    ):
        self._content.set_active_tab("Random")
        if split == "train" or split == "training":
            self._random_splits_table.set_train_split_percent(percent)
        elif split == "val" or split == "validation":
            self._random_splits_table.set_val_split_percent(percent)
        else:
            raise ValueError("Split value must be 'train', 'training', 'val' or 'validation'")

    def get_train_split_percent(self) -> int:
        return self._random_splits_table.get_train_split_percent()

    def get_val_split_percent(self) -> int:
        return 100 - self._random_splits_table.get_train_split_percent()

    def set_tags_splits(
        self, train_tag: str, val_tag: str, untagged_action: Literal["train", "val", "ignore"]
    ):
        self._content.set_active_tab("Based on item tags")
        self._train_tag_select.set_name(train_tag)
        self._val_tag_select.set_name(val_tag)
        self._untagged_select.set_value(untagged_action)

    def get_train_tag(self) -> str:
        return self._train_tag_select.get_selected_name()

    def get_val_tag(self) -> str:
        return self._val_tag_select.get_selected_name()

    def set_datasets_splits(self, train_datasets: List[int], val_datasets: List[int]):
        self._content.set_active_tab("Based on datasets")
        self._train_ds_select.set_dataset_ids(train_datasets)
        self._val_ds_select.set_dataset_ids(val_datasets)

    def get_train_dataset_ids(self) -> List[int]:
        return self._train_ds_select.get_selected_ids()

    def get_val_dataset_ids(self) -> List[int]:
        return self._val_ds_select.get_selected_ids()

    def set_project_id_for_collections(self, project_id: int):
        if not isinstance(project_id, int):
            raise ValueError("Project ID must be an integer.")
        self._project_id = project_id
        self._project_type = None
        if self._api is None:
            self._api = Api()
        self._project_info = self._api.project.get_info_by_id(self._project_id, raise_error=True)
        self._project_type = self._project_info.type
        self._project_class = get_project_class(self._project_type)
        if not self._train_collections_select or not self._val_collections_select:
            raise ValueError("Collections select widgets are not initialized.")
        self._train_collections_select.set_project_id(project_id)
        self._val_collections_select.set_project_id(project_id)

    def get_train_collections_ids(self) -> List[int]:
        return self._train_collections_select.get_selected_ids() or []

    def get_val_collections_ids(self) -> List[int]:
        return self._val_collections_select.get_selected_ids() or []

    def set_collections_splits(self, train_collections: List[int], val_collections: List[int]):
        self._content.set_active_tab("Based on collections")
        self.set_collections_splits_by_ids("train", train_collections)
        self.set_collections_splits_by_ids("val", val_collections)

    def set_collections_splits_by_ids(
        self, split: Literal["train", "val"], collection_ids: List[int]
    ):
        if split == "train":
            self._train_collections_select.set_collections(collection_ids)
        elif split == "val":
            self._val_collections_select.set_collections(collection_ids)
        else:
            raise ValueError("Split value must be 'train' or 'val'")

    def get_untagged_action(self) -> str:
        return self._untagged_select.get_value()

    def disable(self):
        self._content.disable()
        if self._random_splits_table is not None:
            self._random_splits_table.disable()
        if self._train_tag_select is not None:
            self._train_tag_select.disable()
        if self._val_tag_select is not None:
            self._val_tag_select.disable()
        if self._untagged_select is not None:
            self._untagged_select.disable()
        if self._train_ds_select is not None:
            self._train_ds_select.disable()
        if self._val_ds_select is not None:
            self._val_ds_select.disable()
        self._disabled = True
        if self._train_collections_select is not None:
            self._train_collections_select.disable()
        if self._val_collections_select is not None:
            self._val_collections_select.disable()
        DataJson()[self.widget_id]["disabled"] = self._disabled
        DataJson().send_changes()

    def enable(self):
        self._content.enable()
        if self._random_splits_table is not None:
            self._random_splits_table.enable()
        if self._train_tag_select is not None:
            self._train_tag_select.enable()
        if self._val_tag_select is not None:
            self._val_tag_select.enable()
        if self._untagged_select is not None:
            self._untagged_select.enable()
        if self._train_ds_select is not None:
            self._train_ds_select.enable()
        if self._val_ds_select is not None:
            self._val_ds_select.enable()
        self._disabled = False
        if self._train_collections_select is not None:
            self._train_collections_select.enable()
        if self._val_collections_select is not None:
            self._val_collections_select.enable()
        DataJson()[self.widget_id]["disabled"] = self._disabled
        DataJson().send_changes()



import numpy as np

from supervisely import ObjClass, ObjClassCollection, ProjectMeta, logger
from supervisely.geometry.cuboid_3d import Cuboid3d
from supervisely.geometry.point_3d import Vector3d
from supervisely.pointcloud_annotation.pointcloud_figure import PointcloudFigure
from supervisely.pointcloud_annotation.pointcloud_object import PointcloudObject

FOLDER_NAMES = ["velodyne", "image_2", "label_2", "calib"]


def read_kitti_label(label_path, calib_path):
    """
    Read KITTI label file with calibration.
    """

    import open3d as o3d  # pylint: disable=import-error
    try:
        calib = o3d.ml.datasets.KITTI.read_calib(calib_path)
        label = o3d.ml.datasets.KITTI.read_label(label_path, calib)
        return label
    except Exception as e:
        logger.warning(f"Failed to read KITTI label or calibration: {e}")
        return None


def convert_labels_to_meta(labels):
    labels = flatten(labels)
    unique_labels = np.unique([l.label_class for l in labels])
    obj_classes = [ObjClass(k, Cuboid3d) for k in unique_labels]
    meta = ProjectMeta(obj_classes=ObjClassCollection(obj_classes))
    return meta


def convert_bin_to_pcd(src, dst):
    import open3d as o3d  # pylint: disable=import-error

    try:
        bin = np.fromfile(src, dtype=np.float32).reshape(-1, 4)
    except ValueError as e:
        raise Exception(
            f"Incorrect data in the KITTI 3D pointcloud file: {src}. "
            f"There was an error while trying to reshape the data into a 4-column matrix: {e}. "
            "Please ensure that the binary file contains a multiple of 4 elements to be "
            "successfully reshaped into a (N, 4) array.\n"
        )
    points = bin[:, 0:3]
    intensity = bin[:, -1]
    intensity_fake_rgb = np.zeros((intensity.shape[0], 3))
    intensity_fake_rgb[:, 0] = intensity
    pc = o3d.geometry.PointCloud(o3d.utility.Vector3dVector(points))
    pc.colors = o3d.utility.Vector3dVector(intensity_fake_rgb)
    o3d.io.write_point_cloud(dst, pc)


def flatten(list_2d):
    return sum(list_2d, [])


def _convert_label_to_geometry(label):
    geometries = []
    for l in label:
        bbox = l.to_xyzwhlr()
        dim = bbox[[3, 5, 4]]
        pos = bbox[:3] + [0, 0, dim[1] / 2]
        yaw = bbox[-1]
        position = Vector3d(float(pos[0]), float(pos[1]), float(pos[2]))
        rotation = Vector3d(0, 0, float(-yaw))

        dimension = Vector3d(float(dim[0]), float(dim[2]), float(dim[1]))
        geometry = Cuboid3d(position, rotation, dimension)
        geometries.append(geometry)
    return geometries


def convert_label_to_annotation(label, meta, renamed_class_names: dict = None):
    geometries = _convert_label_to_geometry(label)
    figures = []
    objs = []
    for l, geometry in zip(label, geometries):  # by object in point cloud
        class_name = renamed_class_names.get(l.label_class, l.label_class)
        pcobj = PointcloudObject(meta.get_obj_class(class_name))
        figures.append(PointcloudFigure(pcobj, geometry))
        objs.append(pcobj)

    return objs, figures


def convert_calib_to_image_meta(image_name, calib_path, camera_num=2):
    with open(calib_path, "r") as f:
        lines = f.readlines()

    assert 0 < camera_num < 4
    intrinsic_matrix = lines[camera_num].strip().split(" ")[1:]
    intrinsic_matrix = np.array(intrinsic_matrix, dtype=np.float32).reshape(3, 4)[:3, :3]

    obj = lines[4].strip().split(" ")[1:]
    rect_4x4 = np.eye(4, dtype=np.float32)
    rect_4x4[:3, :3] = np.array(obj, dtype=np.float32).reshape(3, 3)

    obj = lines[5].strip().split(" ")[1:]
    Tr_velo_to_cam = np.eye(4, dtype=np.float32)
    Tr_velo_to_cam[:3] = np.array(obj, dtype=np.float32).reshape(3, 4)
    world_cam = np.transpose(rect_4x4 @ Tr_velo_to_cam)
    extrinsic_matrix = world_cam[:4, :3].T

    data = {
        "name": image_name,
        "meta": {
            "deviceId": "CAM_LEFT",
            "sensorsData": {
                "extrinsicMatrix": list(extrinsic_matrix.flatten().astype(float)),
                "intrinsicMatrix": list(intrinsic_matrix.flatten().astype(float)),
            },
        },
    }
    return data



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any, Dict

from supervisely.api.api import Api
from supervisely.api.module_api import ApiField
from supervisely.geometry.mask_3d import Mask3D
from supervisely.project.data_version import VersionSchemaField
from supervisely.project.project_meta import ProjectMeta
from supervisely.project.versioning.schema_fields import VersionSchemaField
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.volume_annotation.volume_annotation import VolumeAnnotation


@dataclass(frozen=True)
class VolumeSnapshotSchema:
    schema_version: str

    def datasets_table_schema(self, pa_module):
        return pa_module.schema(
            [
                (VersionSchemaField.SRC_DATASET_ID, pa_module.int64()),
                (VersionSchemaField.JSON, pa_module.large_string()),
            ]
        )

    def volumes_table_schema(self, pa_module):
        return pa_module.schema(
            [
                (VersionSchemaField.SRC_VOLUME_ID, pa_module.int64()),
                (VersionSchemaField.SRC_DATASET_ID, pa_module.int64()),
                (VersionSchemaField.JSON, pa_module.large_string()),
            ]
        )

    def annotations_table_schema(self, pa_module):
        return pa_module.schema(
            [
                (VersionSchemaField.SRC_VOLUME_ID, pa_module.int64()),
                (VersionSchemaField.ANNOTATION, pa_module.large_string()),
            ]
        )

    def dataset_row_from_record(self, dataset_record: dict) -> Dict[str, Any]:
        return {
            VersionSchemaField.SRC_DATASET_ID: dataset_record.get(ApiField.ID),
            VersionSchemaField.JSON: json.dumps(dataset_record, ensure_ascii=False),
        }

    def volume_row_from_record(self, volume_record: dict) -> Dict[str, Any]:
        return {
            VersionSchemaField.SRC_VOLUME_ID: volume_record.get(ApiField.ID),
            VersionSchemaField.SRC_DATASET_ID: volume_record.get(ApiField.DATASET_ID),
            VersionSchemaField.JSON: json.dumps(volume_record, ensure_ascii=False),
        }

    def annotation_row_from_dict(self, *, src_volume_id: int, annotation: dict) -> Dict[str, Any]:
        return {
            VersionSchemaField.SRC_VOLUME_ID: src_volume_id,
            VersionSchemaField.ANNOTATION: json.dumps(annotation, ensure_ascii=False),
        }

    def annotation_dict_from_raw(
        self,
        *,
        api: Api,
        raw_ann_json: dict,
        project_meta_obj: ProjectMeta,
        key_id_map: KeyIdMap,
    ) -> dict:
        ann = VolumeAnnotation.from_json(raw_ann_json, project_meta_obj, key_id_map)
        self._load_mask_geometries(api, ann, key_id_map)
        return ann.to_json()

    @staticmethod
    def _load_mask_geometries(api: Api, ann: VolumeAnnotation, key_id_map: KeyIdMap) -> None:
        for sf in ann.spatial_figures:
            if sf.geometry.name() != Mask3D.name():
                continue
            api.volume.figure.load_sf_geometry(sf, key_id_map)


_VOLUME_SCHEMAS: Dict[str, VolumeSnapshotSchema] = {
    "v2.0.0": VolumeSnapshotSchema(schema_version="v2.0.0"),
}



from typing import List, Dict, Optional

from supervisely.app.widgets import Widget
from supervisely.app.content import StateJson
from supervisely.api.user_api import UserInfo


class MembersListPreview(Widget):
    def __init__(
        self,
        users: Optional[List[UserInfo]] = [],
        max_width: int = 300,
        empty_text: str = None,
        widget_id: int = None,
    ):
        self._users = users
        self._max_width = self._get_max_width(max_width)
        self._empty_text = empty_text
        super().__init__(widget_id=widget_id, file_path=__file__)

    def _get_max_width(self, value):
        if value < 150:
            value = 150
        return f"{value}px"

    def get_json_data(self):
        return {
            "maxWidth": self._max_width,
        }

    def get_json_state(self) -> Dict:
        return {
            "users": [
                {
                    "login": f"<b>{user.login}</b>",
                    "role": user.role,
                }
                for user in self._users
            ]
        }

    def set(self, users: List[UserInfo]):
        self._users = users
        StateJson()[self.widget_id] = self.get_json_state()
        StateJson().send_changes()
